import json


class ConfigManager:

    server_config = {}  # name, port, ssl_path
    db_config = {}      # name, url, port, username, password
    redis_config = {}

    def __init__(self):
        pass

    def load_config(self, filename):
        try:
            config_file = open(filename, mode='rt', encoding='utf8')
            config_json = json.load(config_file)
            config_file.close()
        except IOError as e:
            print("Error loading file: " + filename, e)
            return

        except json.JSONDecodeError as e:
            print("Error loading json file: ", filename, e)
            return

        if self.validate_server_config(config_json['Server']):
            self.server_config = config_json["Server"]

        if self.validate_database_config(config_json['Database']):
            self.db_config = config_json["Database"]

        if self.validate_redis_config(config_json['Redis']):
            self.redis_config = config_json["Redis"]

    def create_defaults(self):
        # Server fake config
        server_required_fields = ['name', 'port', 'use_ssl', 'ssl_path', 'hostname', 'site_certificate',
                                  'site_private_key', 'ca_certificate', 'ca_private_key', 'upload_path',
                                  'enable_docs']

        for field in server_required_fields:
            self.server_config[field] = ''

        self.server_config['upload_path'] = 'uploads'
        self.server_config['hostname'] = '127.0.0.1'
        self.server_config['port'] = 4040
        self.server_config['enable_docs'] = True

        # Database fake config
        database_required_fields = ['name', 'port', 'url', 'username', 'password']
        for field in database_required_fields:
            self.db_config[field] = ''

        # Redis fake config
        redis_required_fields = ['hostname', 'port', 'db', 'username', 'password']
        for field in redis_required_fields:
            self.redis_config[field] = ''
        # Default redis configuration
        self.redis_config['hostname'] = '127.0.0.1'
        self.redis_config['port'] = 6379
        # Already set to empty string previously
        self.redis_config['username'] = ''
        self.redis_config['password'] = ''
        self.redis_config['db'] = 0

    @staticmethod
    def validate_server_config(config):
        rval = True

        required_fields = ['name', 'port', 'use_ssl', 'ssl_path', 'hostname',
                           'site_certificate', 'site_private_key', 'ca_certificate', 'ca_private_key', 'upload_path']
        for field in required_fields:
            if field not in config:
                print('ERROR: Server Config - missing server ' + field)
                rval = False

        # Add optional debug flag
        if 'debug_mode' not in config:
            config['debug_mode'] = False

        # Add optional enable docs
        if 'enable_docs' not in config:
            config['enable_docs'] = False

        return rval

    @staticmethod
    def validate_database_config(config):
        rval = True
        required_fields = ['name', 'port', 'url', 'username', 'password']
        for field in required_fields:
            if field not in config:
                print('ERROR: Database Config - missing database ' + field)
                rval = False
        return rval

    @staticmethod
    def validate_redis_config(config):
        """
        :param config:
        :return:
        """
        rval = True
        required_fields = ['hostname', 'port', 'db', 'username', 'password']
        for field in required_fields:
            if field not in config:
                print('ERROR: Redis Config - missing database ' + field)
                rval = False
        return rval

    @staticmethod
    def validate_service_config(service_name, config):
        """
       :param config:
       :return:
       """
        rval = True
        # TODO make sure to update the fields
        required_fields = ['ServiceUUID', 'ServiceHostname', 'ServicePort', 'ServiceEndpoint', 'ClientEndpoint']
        for field in required_fields:
            if field not in config:
                print('ERROR: Service Config - missing config field  in service', service_name, ' field:', field)
                rval = False
        return rval



from opentera.db.Base import BaseModel
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError
from opentera.db.models.TeraDevice import TeraDevice


class TeraDeviceSubType(BaseModel):

    __tablename__ = 't_devices_subtypes'
    id_device_subtype = Column(Integer, Sequence('id_device_subtype_sequence'), primary_key=True, autoincrement=True)
    id_device_type = Column(Integer, ForeignKey('t_devices_types.id_device_type', ondelete='cascade'), nullable=False)
    device_subtype_name = Column(String, nullable=False)

    device_subtype_type = relationship("TeraDeviceType")

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields += ['device_subtype_type']

        device_subtype_json = super().to_json(ignore_fields=ignore_fields)

        # Add device type name as parent
        device_subtype_json['device_subtype_parent'] = self.device_subtype_type.device_type_name

        return device_subtype_json

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraDeviceType import TeraDeviceType
            bureau = TeraDeviceType.get_device_type_by_key('bureau_actif')
            subtype = TeraDeviceSubType()
            subtype.device_subtype_type = bureau
            subtype.device_subtype_name = 'Bureau modèle #1'
            TeraDeviceSubType.db().session.add(subtype)

            subtype = TeraDeviceSubType()
            subtype.device_subtype_type = bureau
            subtype.device_subtype_name = 'Bureau modèle #2'
            TeraDeviceSubType.db().session.add(subtype)

            TeraDeviceSubType.db().session.commit()

    @staticmethod
    def get_devices_subtypes():
        return TeraDeviceSubType.query.all()

    @staticmethod
    def get_device_subtype_by_name(name: str, id_device_type: int):
        return (TeraDeviceSubType.query.filter_by(device_subtype_name=name).filter_by(id_device_type=id_device_type)
                .first())

    @staticmethod
    def get_device_subtype_by_id(dev_subtype: int):
        return TeraDeviceSubType.query.filter_by(id_device_subtype=dev_subtype).first()

    @staticmethod
    def get_device_subtypes_for_type(dev_type: int):
        return TeraDeviceSubType.query.filter_by(id_device_type=dev_type).all()

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        if (TeraDevice.get_count(filters={'id_device_subtype': self.id_device_subtype}, with_deleted=with_deleted)) > 0:
            return IntegrityError('Device subtype still have devices with that subtype', self.id_device_subtype,
                                  't_devices')
        return None



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from opentera.db.models.TeraSessionType import TeraSessionType
from opentera.db.models.TeraSessionTypeProject import TeraSessionTypeProject
from opentera.db.models.TeraTestTypeProject import TeraTestTypeProject
from opentera.db.models.TeraTestType import TeraTestType
from opentera.db.models.TeraTest import TeraTest
from opentera.db.models.TeraSession import TeraSession
from opentera.db.models.TeraAsset import TeraAsset
from sqlalchemy import Column, ForeignKey, Integer, Sequence, or_
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraServiceProject(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_services_projects'
    id_service_project = Column(Integer, Sequence('id_service_project_sequence'), primary_key=True, autoincrement=True)
    id_service = Column(Integer, ForeignKey('t_services.id_service', ondelete='cascade'), nullable=False)
    id_project = Column(Integer, ForeignKey('t_projects.id_project', ondelete='cascade'), nullable=False)

    service_project_service = relationship("TeraService", viewonly=True)
    service_project_project = relationship("TeraProject", viewonly=True)

    def __init__(self):
        pass

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_project_service', 'service_project_project'])

        if minimal:
            ignore_fields.extend([])

        return super().to_json(ignore_fields=ignore_fields)

    @staticmethod
    def get_services_for_project(id_project: int, with_deleted: bool = False):
        return TeraServiceProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_project=id_project).all()

    @staticmethod
    def get_projects_for_service(id_service: int, with_deleted: bool = False):
        return TeraServiceProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=id_service).all()

    @staticmethod
    def get_service_project_by_id(service_project_id: int, with_deleted: bool = False):
        return TeraServiceProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service_project=service_project_id).first()

    @staticmethod
    def get_service_project_for_service_project(project_id: int, service_id: int, with_deleted: bool = False):
        return TeraServiceProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_project=project_id, id_service=service_id).first()

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraService import TeraService
            from opentera.db.models.TeraProject import TeraProject

            project1 = TeraProject.get_project_by_projectname('Default Project #1')
            project2 = TeraProject.get_project_by_projectname('Default Project #2')

            servicefile = TeraService.get_service_by_key('FileTransferService')
            servicevideorehab = TeraService.get_service_by_key('VideoRehabService')

            service_project = TeraServiceProject()
            service_project.id_project = project1.id_project
            service_project.id_service = servicefile.id_service
            TeraServiceProject.db().session.add(service_project)

            service_project = TeraServiceProject()
            service_project.id_project = project1.id_project
            service_project.id_service = servicevideorehab.id_service
            TeraServiceProject.db().session.add(service_project)

            service_project = TeraServiceProject()
            service_project.id_project = project2.id_project
            service_project.id_service = servicefile.id_service
            TeraServiceProject.db().session.add(service_project)

            service_project = TeraServiceProject()
            service_project.id_project = 3
            service_project.id_service = servicefile.id_service
            TeraServiceProject.db().session.add(service_project)

            TeraServiceProject.db().session.commit()

    @classmethod
    def insert(cls, stp):
        # Check if that site of that project has the site associated to it
        from opentera.db.models.TeraServiceSite import TeraServiceSite
        from opentera.db.models.TeraProject import TeraProject
        project = TeraProject.get_project_by_id(project_id=stp.id_project)
        service_site = TeraServiceSite.get_service_site_for_service_site(site_id=project.id_site,
                                                                         service_id=stp.id_service)
        if not service_site:
            raise IntegrityError(params='Service not associated to project site', orig='TeraServiceProject.insert',
                                 statement='insert')
        return super().insert(stp)

    @classmethod
    def update(cls, update_id: int, values: dict):
        return

    @staticmethod
    def delete_with_ids(service_id: int, project_id: int, autocommit: bool = True):
        delete_obj = TeraServiceProject.query.filter_by(id_service=service_id, id_project=project_id).first()
        if delete_obj:
            TeraServiceProject.delete(delete_obj.id_service_project, autocommit=autocommit)

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        # This check will be quite long to process with lot of sessions and data...
        session_types_ids = \
            [st.id_session_type for st in TeraSessionType.get_session_types_for_service(self.id_service,
                                                                                        with_deleted=with_deleted)]

        sessions = TeraSession.get_sessions_for_project(self.id_project, with_deleted=with_deleted)
        for session in sessions:
            if session.id_session_type in session_types_ids:
                return IntegrityError('Service has sessions of related session type in this project', self.id_service,
                                      't_sessions')
            if TeraTest.query.join(TeraTestType).filter(TeraTest.id_session == session.id_session).\
                    filter(or_(TeraTest.id_service == self.id_service, TeraTestType.id_service == self.id_service))\
                    .execution_options(include_deleted=with_deleted).count() > 0:
                return IntegrityError('Service has tests of related test type in this project', self.id_service,
                                      't_tests')
            if TeraAsset.query.filter_by(id_session=session.id_session).\
                filter(or_(TeraAsset.asset_service_uuid == self.service_project_service.service_uuid,
                           TeraAsset.id_service == self.id_service)).execution_options(include_deleted=with_deleted)\
                    .count() > 0:
                return IntegrityError('Service has related assets in this project', self.id_service, 't_assets')
        return None

    @classmethod
    def delete(cls, id_todel, autocommit: bool = True):
        # Delete all session type association to that project
        delete_obj: TeraServiceProject = TeraServiceProject.query.filter_by(id_service_project=id_todel).first()

        if delete_obj:
            session_types = TeraSessionTypeProject.get_session_type_project_for_project_and_service(
                project_id=delete_obj.id_project, service_id=delete_obj.id_service)
            for session_type in session_types:
                TeraSessionTypeProject.delete(session_type.id_session_type_project, autocommit=autocommit)

            test_types = TeraTestTypeProject.get_test_type_project_for_project_and_service(
                project_id=delete_obj.id_project, service_id=delete_obj.id_service)
            for test_type in test_types:
                TeraTestTypeProject.delete(test_type.id_test_type_project, autocommit=autocommit)

            # Ok, delete it
            super().delete(id_todel, autocommit=autocommit)



from opentera.forms.TeraForm import *

from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from flask_babel import gettext


class TeraDeviceTypeForm:

    @staticmethod
    def get_device_type_form():
        form = TeraForm("device_type")

        # Building lists
        #################

        # Sections
        section = TeraFormSection("infos", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_device_type", gettext("Device Type ID"), "hidden",
                                      item_required=True))
        section.add_item(TeraFormItem("device_type_name", gettext("Device Type Name"), "text",
                                      item_required=True))
        section.add_item(TeraFormItem("device_type_key", gettext("Device Type Key"), "text",
                                      item_required=True))

        return form.to_dict()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: LoginEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import LogEvent_pb2 as LogEvent__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10LoginEvent.proto\x12\x11opentera.protobuf\x1a\x0eLogEvent.proto\"\x89\x07\n\nLoginEvent\x12/\n\nlog_header\x18\x01 \x01(\x0b\x32\x1b.opentera.protobuf.LogEvent\x12;\n\nlogin_type\x18\x02 \x01(\x0e\x32\'.opentera.protobuf.LoginEvent.LoginType\x12?\n\x0clogin_status\x18\x03 \x01(\x0e\x32).opentera.protobuf.LoginEvent.LoginStatus\x12\x13\n\x0b\x63lient_name\x18\x04 \x01(\t\x12\x16\n\x0e\x63lient_version\x18\x05 \x01(\t\x12\x11\n\tclient_ip\x18\x06 \x01(\t\x12\x0f\n\x07os_name\x18\x07 \x01(\t\x12\x12\n\nos_version\x18\x08 \x01(\t\x12\x11\n\tuser_uuid\x18\t \x01(\t\x12\x18\n\x10participant_uuid\x18\n \x01(\t\x12\x13\n\x0b\x64\x65vice_uuid\x18\x0b \x01(\t\x12\x14\n\x0cservice_uuid\x18\x0c \x01(\t\x12\x17\n\x0fserver_endpoint\x18\r \x01(\t\"\x85\x03\n\x0bLoginStatus\x12\x18\n\x14LOGIN_STATUS_UNKNOWN\x10\x00\x12\x18\n\x14LOGIN_STATUS_SUCCESS\x10\x01\x12+\n\'LOGIN_STATUS_FAILED_WITH_WRONG_PASSWORD\x10\x02\x12+\n\'LOGIN_STATUS_FAILED_WITH_WRONG_USERNAME\x10\x03\x12\x31\n-LOGIN_STATUS_FAILED_WITH_MAX_ATTEMPTS_REACHED\x10\x04\x12.\n*LOGIN_STATUS_FAILED_WITH_ALREADY_LOGGED_IN\x10\x05\x12*\n&LOGIN_STATUS_FAILED_WITH_INVALID_TOKEN\x10\x06\x12-\n)LOGIN_STATUS_FAILED_WITH_DISABLED_ACCOUNT\x10\x07\x12*\n&LOGIN_STATUS_FAILED_WITH_EXPIRED_TOKEN\x10\x08\"n\n\tLoginType\x12\x16\n\x12LOGIN_TYPE_UNKNOWN\x10\x00\x12\x14\n\x10LOGIN_TYPE_TOKEN\x10\x01\x12\x17\n\x13LOGIN_TYPE_PASSWORD\x10\x02\x12\x1a\n\x16LOGIN_TYPE_CERTIFICATE\x10\x03\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'LoginEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _LOGINEVENT._serialized_start=56
  _LOGINEVENT._serialized_end=961
  _LOGINEVENT_LOGINSTATUS._serialized_start=460
  _LOGINEVENT_LOGINSTATUS._serialized_end=849
  _LOGINEVENT_LOGINTYPE._serialized_start=851
  _LOGINEVENT_LOGINTYPE._serialized_end=961
# @@protoc_insertion_point(module_scope)



from redis import Redis
import opentera.messages.python as messages
import datetime


class LoggingClient:
    """
    enum LogLevel
    {
        LOGLEVEL_UNKNOWN = 0;
        LOGLEVEL_TRACE = 1;
        LOGLEVEL_DEBUG = 2;
        LOGLEVEL_INFO = 3;
        LOGLEVEL_WARNING = 4;
        LOGLEVEL_CRITICAL = 5;
        LOGLEVEL_ERROR = 6;
        LOGLEVEL_FATAL = 7;

        LogLevel level = 1;
        double timestamp = 2;
        string sender = 3;
        string message = 4;
    }

    This client will send messages to queues. Logging Service is responsible of reading the queues.
    """

    def __init__(self, config, client_name='LoggingClient'):
        # Setup connection with redis
        self.client_name = client_name
        self.redis = Redis(host=config['hostname'],
                           port=config['port'],
                           username=config['username'],
                           password=config['password'],
                           db=config['db'],
                           client_name=self.client_name)

    def close(self):
        print('Closing :', self.client_name)
        self.redis.close()

    def log_trace(self, sender: str, *args):
        try:
            self._push_tera_event_message(
                LoggingClient._create_log_event(messages.LogEvent.LOGLEVEL_TRACE, sender, *args))
        # TODO Can we do better than catch base Exception here
        except Exception as e:
            print('LoggingClient.log_trace', e)

    def log_debug(self, sender: str, *args):
        try:
            self._push_tera_event_message(
                LoggingClient._create_log_event(messages.LogEvent.LOGLEVEL_DEBUG, sender, *args))
        # TODO Can we do better than catch base Exception here
        except Exception as e:
            print('LoggingClient.log_debug', e)

    def log_info(self, sender: str, *args):
        try:
            self._push_tera_event_message(
                LoggingClient._create_log_event(messages.LogEvent.LOGLEVEL_INFO, sender, *args))
        # TODO Can we do better than catch base Exception here
        except Exception as e:
            print(self.client_name + '.log_info', e)

    def log_warning(self, sender: str, *args):
        try:
            self._push_tera_event_message(
                LoggingClient._create_log_event(messages.LogEvent.LOGLEVEL_WARNING, sender, *args))
        # TODO Can we do better than catch base Exception here
        except Exception as e:
            print(self.client_name + '.log_warning', e)

    def log_critical(self, sender: str, *args):
        try:
            self._push_tera_event_message(
                LoggingClient._create_log_event(messages.LogEvent.LOGLEVEL_CRITICAL, sender, *args))
        # TODO Can we do better than catch base Exception here
        except Exception as e:
            print(self.client_name + '.log_critical', e)

    def log_error(self, sender: str, *args):
        try:
            self._push_tera_event_message(
                LoggingClient._create_log_event(messages.LogEvent.LOGLEVEL_ERROR, sender, *args))
        # TODO Can we do better than catch base Exception here
        except Exception as e:
            print(self.client_name + '.log_error', e)

    def log_fatal(self, sender: str, *args):
        try:
            self._push_tera_event_message(
                LoggingClient._create_log_event(messages.LogEvent.LOGLEVEL_FATAL, sender, *args))
        # TODO Can we do better than catch base Exception here
        except Exception as e:
            print(self.client_name + '.log_fatal', e)

    def send_login_event(self, sender: str, level: int, login_type: int, login_status: int,
                         client_name: str, client_version: str, client_ip: str, os_name: str, os_version: str,
                         user_uuid: str = '', participant_uuid: str = '', device_uuid: str = '', service_uuid: str = '',
                         server_endpoint: str = '', message: str = ''):
        # Create and fill event info
        login_event = messages.LoginEvent()
        login_event.log_header.level = level
        login_event.log_header.timestamp = datetime.datetime.now().timestamp()
        login_event.log_header.sender = str(sender)
        login_event.log_header.message = message
        login_event.login_type = login_type
        login_event.login_status = login_status
        login_event.client_name = client_name
        login_event.client_version = client_version
        login_event.client_ip = client_ip
        login_event.os_name = os_name
        login_event.os_version = os_version
        login_event.user_uuid = '' if user_uuid is None else user_uuid
        login_event.participant_uuid = '' if participant_uuid is None else participant_uuid
        login_event.device_uuid = '' if device_uuid is None else device_uuid
        login_event.service_uuid = '' if service_uuid is None else service_uuid
        login_event.server_endpoint = server_endpoint
        # Send it
        try:
            self._push_tera_login_event_message(login_event)
        except Exception as e:
            print(self.client_name + '.' + LoggingClient._loglevel_to_topic_string(login_event.log_header.level), e)

    @staticmethod
    def _create_log_event(level, sender, *args):
        log_event = messages.LogEvent()
        log_event.level = level
        log_event.timestamp = datetime.datetime.now().timestamp()
        log_event.sender = str(sender)
        log_event.message = str(args)
        return log_event

    @staticmethod
    def _loglevel_to_topic_string(level: int) -> str:
        level_string = 'unknown'

        # Convert level to string
        if level == messages.LogEvent.LOGLEVEL_TRACE:
            level_string = 'trace'
        elif level == messages.LogEvent.LOGLEVEL_DEBUG:
            level_string = 'debug'
        elif level == messages.LogEvent.LOGLEVEL_INFO:
            level_string = 'info'
        elif level == messages.LogEvent.LOGLEVEL_WARNING:
            level_string = 'warning'
        elif level == messages.LogEvent.LOGLEVEL_CRITICAL:
            level_string = 'critical'
        elif level == messages.LogEvent.LOGLEVEL_ERROR:
            level_string = 'error'
        elif level == messages.LogEvent.LOGLEVEL_FATAL:
            level_string = 'fatal'

        topic = 'log.' + level_string
        return topic

    @staticmethod
    def _create_tera_event_message(log_event: messages.LogEvent):
        event_message = messages.TeraEvent()
        event_message.header.version = 1
        # Duplicated time, useful?
        event_message.header.time = log_event.timestamp
        event_message.header.topic = LoggingClient._loglevel_to_topic_string(log_event.level)

        # Add log event
        any_message = messages.Any()
        any_message.Pack(log_event)
        event_message.events.extend([any_message])

        return event_message

    @staticmethod
    def _create_tera_event_message_from_login_event(login_event: messages.LoginEvent):
        event_message = messages.TeraEvent()
        event_message.header.version = 1
        # Duplicated time, useful?
        event_message.header.time = login_event.log_header.timestamp
        event_message.header.topic = LoggingClient._loglevel_to_topic_string(login_event.log_header.level)

        # Add log event
        any_message = messages.Any()
        any_message.Pack(login_event)
        event_message.events.extend([any_message])

        return event_message

    def _push_tera_event_message(self, log_event: messages.LogEvent):
        event_message = LoggingClient._create_tera_event_message(log_event)
        self.redis.rpush(event_message.header.topic, event_message.SerializeToString())
        return event_message

    def _push_tera_login_event_message(self, login_event: messages.LoginEvent):
        event_message = LoggingClient._create_tera_event_message_from_login_event(login_event)
        self.redis.rpush(event_message.header.topic, event_message.SerializeToString())
        return event_message


if __name__ == '__main__':
    from opentera.config.ConfigManager import ConfigManager
    import uuid

    config_man = ConfigManager()
    config_man.create_defaults()

    # Create logger
    client = LoggingClient(config_man.redis_config)
    client.log_trace('TRACE-TEST', 'hello world', 1, 2, 3, 'test', client)
    client.log_debug('DEBUG-TEST', 'hello world', 1, 2, 3, 'test', client)
    client.log_info('INFO-TEST', 'hello world', 1, 2, 3, 'test', client)
    client.log_warning('WARNING-TEST', 'hello world', 1, 2, 3, 'test', client)
    client.log_error('ERROR-TEST', 'hello world', 1, 2, 3, 'test', client)
    client.log_critical('CRITICAL-TEST', 'hello world', 1, 2, 3, 'test', client)
    client.log_fatal('FATAL-TEST', 'hello world', 1, 2, 3, 'test', client)

    client.send_login_event(sender='unknown',
                            level=messages.LogEvent.LOGLEVEL_INFO,
                            login_type=messages.LoginEvent.LOGIN_TYPE_TOKEN,
                            login_status=messages.LoginEvent.LOGIN_STATUS_SUCCESS,
                            client_name='test',
                            client_version="1.0",
                            client_ip='127.0.0.1',
                            os_name='Unknown',
                            os_version="Unknown",
                            user_uuid=str(uuid.uuid4()),
                            participant_uuid=str(uuid.uuid4()),
                            device_uuid=str(uuid.uuid4()),
                            service_uuid=str(uuid.uuid4()),
                            server_endpoint='/')




# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: JoinSessionEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16JoinSessionEvent.proto\x12\x11opentera.protobuf\"\xed\x01\n\x10JoinSessionEvent\x12\x13\n\x0bsession_url\x18\x01 \x01(\t\x12\x1c\n\x14session_creator_name\x18\x02 \x01(\t\x12\x14\n\x0csession_uuid\x18\x03 \x01(\t\x12\x1c\n\x14session_participants\x18\x04 \x03(\t\x12\x15\n\rsession_users\x18\x05 \x03(\t\x12\x17\n\x0fsession_devices\x18\x06 \x03(\t\x12\x10\n\x08join_msg\x18\x07 \x01(\t\x12\x1a\n\x12session_parameters\x18\x08 \x01(\t\x12\x14\n\x0cservice_uuid\x18\t \x01(\tb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'JoinSessionEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _JOINSESSIONEVENT._serialized_start=46
  _JOINSESSIONEVENT._serialized_end=283
# @@protoc_insertion_point(module_scope)



"""
Functions related to dynamic generation of the soft-delete mixin.
Adapted from:
https://github.com/flipbit03/sqlalchemy-easy-softdelete
"""

from datetime import datetime
from typing import Any, Callable, Optional, Type

from sqlalchemy import Column, DateTime, text
from sqlalchemy.inspection import inspect
from sqlalchemy.sql.type_api import TypeEngine
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.event import listens_for
from sqlalchemy.orm import ORMExecuteState, Session
from sqlalchemy.exc import IntegrityError

from functools import cache

from opentera.db.SoftDeleteQueryRewriter import SoftDeleteQueryRewriter


@cache
def activate_soft_delete_hook(deleted_field_name: str, disable_soft_delete_option_name: str):
    """Activate an event hook to rewrite the queries."""
    # Enable Soft Delete on all Relationship Loads which implement SoftDeleteMixin
    @listens_for(Session, "do_orm_execute")
    def soft_delete_execute(state: ORMExecuteState):
        if not state.is_select:
            return

        if disable_soft_delete_option_name in state.execution_options \
                and state.execution_options[disable_soft_delete_option_name]:
            return

        if 'include_deleted' in state.session.info and len(state.session.info['include_deleted']) > 0:
            return

        adapted = SoftDeleteQueryRewriter(deleted_field_name, disable_soft_delete_option_name).rewrite_statement(
            state.statement
        )
        state.statement = adapted
    # @listens_for(Engine, "before_execute", retval=True)
    # def soft_delete_execute(conn: Connection, clauseelement, multiparams, params, execution_options):
    #     if not isinstance(clauseelement, Select):
    #         return clauseelement, multiparams, params
    #
    #     if disable_soft_delete_option_name in execution_options and execution_options[disable_soft_delete_option_name]:
    #         # print('test_include_deleted')
    #         return clauseelement, multiparams, params
    #
    #     adapted = SoftDeleteQueryRewriter(deleted_field_name, disable_soft_delete_option_name).rewrite_statement(
    #         clauseelement
    #     )
    #     return adapted, multiparams, params


def generate_soft_delete_mixin_class(
    deleted_field_name: str = "deleted_at",
    class_name: str = "_SoftDeleteMixin",
    deleted_field_type: TypeEngine = DateTime(timezone=True),
    disable_soft_delete_filtering_option_name: str = "include_deleted",
    generate_delete_method: bool = True,
    delete_method_name: str = "delete",
    delete_method_default_value: Callable[[], Any] = lambda: datetime.utcnow(),
    generate_undelete_method: bool = True,
    undelete_method_name: str = "undelete",
    handle_cascade_delete: bool = True,
    generate_hard_delete_method: bool = True,
    hard_delete_method_name: str = "hard_delete"
) -> Type:
    """Generate the actual soft-delete Mixin class."""
    class_attributes = {deleted_field_name: Column(deleted_field_name, deleted_field_type)}

    def get_class_from_tablename(_self, tablename: str) -> DeclarativeMeta | None:
        for mapper in _self.registry.mappers:
            if mapper.class_.__tablename__ == tablename:
                return mapper.class_
        return None

    class_attributes['get_class_from_tablename'] = get_class_from_tablename

    if generate_delete_method:
        def delete_method(_self, v: Optional[Any] = None):
            setattr(_self, deleted_field_name, v or delete_method_default_value())
            if handle_cascade_delete:
                primary_key_name = inspect(_self.__class__).primary_key[0].name
                for relation in inspect(_self.__class__).relationships.items():
                    # Relationship has a cascade delete or a secondary table
                    if relation[1].cascade.delete:
                        # Item has a delete_at field (thus supports soft-delete)
                        if deleted_field_name in relation[1].entity.columns.keys():
                            # Cascade soft delete for each item
                            for item in getattr(_self, relation[0]):
                                item_deleter = getattr(item, delete_method_name)
                                item_deleter()
                    if relation[1].secondary is not None:
                        # Item has a delete_at field (thus supports soft-delete)
                        if deleted_field_name in relation[1].entity.columns.keys():
                            model_class = _self.get_class_from_tablename(relation[1].secondary.name)
                            if model_class:
                                related_items = model_class.query.filter(text(primary_key_name + "=" +
                                                                              str(getattr(_self, primary_key_name)))
                                                                         ).all()
                                for item in related_items:
                                    item_deleter = getattr(item, delete_method_name)
                                    item_deleter()

        class_attributes[delete_method_name] = delete_method

    if generate_hard_delete_method:
        def hard_delete_method(_self):
            _self.handle_include_deleted_flag(True)
            # Callback actions before doing hard delete, if required
            if getattr(_self, 'hard_delete_before', None):
                _self.hard_delete_before()
            if handle_cascade_delete:
                primary_key_name = inspect(_self.__class__).primary_key[0].name
                for relation in inspect(_self.__class__).relationships.items():
                    # Relationship has a cascade delete or a secondary table
                    if relation[1].cascade.delete:
                        for item in getattr(_self, relation[0]):
                            # print("Cascade deleting " + str(item))
                            hard_item_deleter = getattr(item, hard_delete_method_name)
                            hard_item_deleter()

                    if relation[1].secondary is not None and relation[1].passive_deletes:
                        if deleted_field_name in relation[1].entity.columns.keys():
                            model_class = _self.get_class_from_tablename(relation[1].secondary.name)
                            if model_class:
                                related_items = model_class.query.filter(text(primary_key_name + "=" +
                                                                              str(getattr(_self, primary_key_name)))
                                                                         ).execution_options(include_deleted=True).all()
                                for item in related_items:
                                    # print("Cascade deleting " + str(model_class) + ": " + primary_key_name + " = " +
                                    #       str(getattr(_self, primary_key_name)))
                                    item_hard_deleter = getattr(item, hard_delete_method_name)
                                    item_hard_deleter()

            if _self not in _self.db().session.deleted:
                _self.db().session.delete(_self)
                _self.commit()
            _self.handle_include_deleted_flag(False)

        class_attributes[hard_delete_method_name] = hard_delete_method

    if generate_undelete_method:
        def get_undelete_cascade_relations(_self) -> list:
            return []  # By default, no relationships are automatically undeleted when undeleting

        class_attributes['get_undelete_cascade_relations'] = get_undelete_cascade_relations

        def undelete_method(_self):
            if not getattr(_self, deleted_field_name):
                print("Object " + str(_self.__class__) + " not deleted - returning.")
                return

            current_obj = inspect(_self.__class__)
            primary_key_name = current_obj.primary_key[0].name

            # Check data integrity before undeleting
            for col in current_obj.columns:
                if not col.foreign_keys:
                    continue
                # If column foreign key is not nullable or there is a value for this object, we must check if
                # related object is still there
                col_value = getattr(_self, col.name, None)
                if not col.nullable or col_value:
                    remote_table_name = list(col.foreign_keys)[0].column.table.name
                    model_class = _self.get_class_from_tablename(remote_table_name)
                    remote_key_name = list(col.foreign_keys)[0].column.key
                    related_item = model_class.query.filter(text(remote_key_name + '="' + str(col_value) + '"')).first()
                    if not related_item:
                        # A required object isn't there (soft-deleted) - throw exception to undelete it first!
                        raise IntegrityError('Cannot undelete: unsatisfied foreign key - ' + col.name, col_value,
                                             remote_table_name)
            # Undelete!
            # print("Undeleting " + str(_self.__class__))
            setattr(_self, deleted_field_name, None)

            # Check relationships that are cascade deleted to restore them
            if handle_cascade_delete:
                for relation in inspect(_self.__class__).relationships.items():
                    # Relationship has a cascade undelete ?
                    if relation[1].key in _self.get_undelete_cascade_relations():
                        # Item has a delete_at field (thus supports soft-delete)
                        if deleted_field_name in relation[1].entity.columns.keys():
                            # Cascade undelete - must manually query to get deleted rows
                            remote_primary_key = list(relation[1].remote_side)[0].name
                            local_primary_key = list(relation[1].local_columns)[0].name
                            self_key_value = getattr(_self, local_primary_key)
                            if not self_key_value:
                                continue
                            related_items = relation[1].entity.class_.query.execution_options(include_deleted=True).\
                                filter(text(remote_primary_key + '=' + str(self_key_value))).all()
                            for item in related_items:
                                # print("--> Undeleting relationship " + relation[1].key)
                                item_undeleter = getattr(item, undelete_method_name)
                                item_undeleter()
                            continue

                    # Check secondary relationships and restore them if both ends are now undeleted
                    if relation[1].secondary is not None:
                        if deleted_field_name in relation[1].entity.columns.keys():
                            model_class = _self.get_class_from_tablename(relation[1].secondary.name)
                            related_items = model_class.query.execution_options(include_deleted=True).\
                                filter(text(primary_key_name + '=' + str(getattr(_self, primary_key_name)))).all()
                            for item in related_items:
                                # Check if other side of the relationship is present and, if so, restores it
                                remote_primary_key = relation[1].target.primary_key.columns[0].name
                                remote_model = _self.get_class_from_tablename(relation[1].target.name)
                                remote_item = remote_model.query.filter(
                                    text(remote_primary_key + '=' + str(getattr(item, remote_primary_key)))).first()
                                if remote_item:
                                    #  print("--> Undeleting relationship with " + relation[1].target.name)
                                    item_undeleter = getattr(item, undelete_method_name)
                                    item_undeleter()

            # _self.handle_include_deleted_flag(True)
            # setattr(_self, deleted_field_name, None)
            # print("Undeleting " + str(_self.__class__))
            # if handle_cascade_delete:
            #     primary_key_name = inspect(_self.__class__).primary_key[0].name
            #     for relation in inspect(_self.__class__).relationships.items():
            #         print(str(_self.__class__) + " - relation " + str(relation))
            #         if relation[1].secondary is not None:
            #             print("-> Undeleting secondary table relationship " + relation[1].secondary.name)
            #             # Item has a delete_at field (thus supports soft-delete)
            #             if deleted_field_name in relation[1].entity.columns.keys():
            #                 model_class = _self.get_class_from_tablename(relation[1].secondary.name)
            #                 if model_class:
            #                     related_items = model_class.query.filter(text(primary_key_name + '=' +
            #                                                                   str(getattr(_self, primary_key_name)))
            #                                                              ).execution_options(include_deleted=True).all()
            #                     for item in related_items:
            #                         item_undeleter = getattr(item, undelete_method_name)
            #                         item_undeleter()
            #
            #                         # Undelete "left-side" item of the relationship
            #                         remote_primary_key = relation[1].target.primary_key.columns[0].name
            #                         remote_model = _self.get_class_from_tablename(relation[1].target.name)
            #                         remote_item = remote_model.query.filter(text(remote_primary_key + '=' +
            #                                                                      str(getattr(item, remote_primary_key)))
            #                                                                 ).execution_options(include_deleted=True)\
            #                             .first()
            #                         if remote_item:
            #                             print("--> Undeleting left side of secondary table " + relation[1].target.name)
            #                             item_undeleter = getattr(remote_item, undelete_method_name)
            #                             item_undeleter()
            #
            #                     continue
            #         # Check for parents or related items
            #         if relation[1].back_populates:
            #             print("--> Undeleting back_populates relationship " + str(relation[1]))
            #             # if relation[1].cascade.delete:  # Relationship has a cascade delete
            #             # Item has a delete_at field (thus supports soft-delete)
            #             if deleted_field_name in relation[1].entity.columns.keys():
            #                 # Cascade undelete - must manually query to get deleted rows
            #                 remote_primary_key = list(relation[1].remote_side)[0].name
            #                 local_primary_key = list(relation[1].local_columns)[0].name
            #                 self_key_value = getattr(_self, local_primary_key)
            #                 if not self_key_value:
            #                     continue
            #                 related_items = relation[1].entity.class_.query.execution_options(include_deleted=True).\
            #                     filter(text(remote_primary_key + '=' + str(self_key_value))).all()
            #                 for item in related_items:
            #                     item_undeleter = getattr(item, undelete_method_name)
            #                     item_undeleter()
            #                 continue
            #         print("Skipped undelete")
            #     _self.handle_include_deleted_flag(False)

        class_attributes[undelete_method_name] = undelete_method

    activate_soft_delete_hook(deleted_field_name, disable_soft_delete_filtering_option_name)

    generated_class = type(class_name, tuple(), class_attributes)

    return generated_class


# Create a Class that inherits from our class builder
class SoftDeleteMixin(generate_soft_delete_mixin_class(delete_method_name='soft_delete',
                                                       undelete_method_name='soft_undelete')):
    # type hint for autocomplete IDE support
    deleted_at: datetime





from opentera.services.ServiceOpenTera import ServiceOpenTera
from opentera.services.ServiceAccessManager import ServiceAccessManager, current_service_client, \
    current_login_type, current_user_client, current_device_client, current_participant_client, LoginType
from opentera.services.ServiceConfigManager import ServiceConfigManager
from opentera.modules.BaseModule import ModuleNames, create_module_event_topic_from_name

import opentera.messages.python as messages
from google.protobuf.json_format import ParseError
from google.protobuf.message import DecodeError
from twisted.internet import defer

import jwt
from abc import ABC, abstractmethod


class ServiceOpenTeraWithTests(ServiceOpenTera, ABC):

    def __init__(self, config_man: ServiceConfigManager, this_service_info):
        ServiceOpenTera.__init__(self, config_man, this_service_info)

    @staticmethod
    def has_access_to_test(access_token: str, test_uuid: str) -> bool:
        try:
            access = jwt.decode(access_token, ServiceAccessManager.api_service_token_key, algorithms='HS256')
        except jwt.PyJWTError:
            return False

        if 'test_uuids' not in access or 'requester_uuid' not in access:
            return False

        requester_uuid = ServiceOpenTeraWithTests.get_current_requester_uuid()
        if access['requester_uuid'] != requester_uuid:
            return False

        if test_uuid not in access['test_uuids']:
            return False

        return True

    @staticmethod
    def has_access_to_test_type(access_token: str, test_type_uuid: str) -> bool:
        try:
            access = jwt.decode(access_token, ServiceAccessManager.api_service_token_key, algorithms='HS256')
        except jwt.PyJWTError:
            return False

        if 'test_type_uuids' not in access or 'requester_uuid' not in access:
            return False

        requester_uuid = ServiceOpenTeraWithTests.get_current_requester_uuid()
        if access['requester_uuid'] != requester_uuid:
            return False

        if test_type_uuid not in access['test_type_uuids']:
            return False

        return True

    @staticmethod
    def can_edit_test_type(access_token) -> bool:
        try:
            access = jwt.decode(access_token, ServiceAccessManager.api_service_token_key, algorithms='HS256')
        except jwt.PyJWTError:
            return False

        if 'can_edit' not in access:
            return False

        return access['can_edit']

    @staticmethod
    def get_accessible_test_uuids(access_token: str) -> list:
        try:
            access = jwt.decode(access_token, ServiceAccessManager.api_service_token_key, algorithms='HS256')
        except jwt.PyJWTError:
            return []

        if 'test_uuids' not in access or 'requester_uuid' not in access:
            return []

        requester_uuid = ServiceOpenTeraWithTests.get_current_requester_uuid()
        if access['requester_uuid'] != requester_uuid:
            return []

        return access['test_uuids']

    @staticmethod
    def get_current_requester_uuid() -> str:
        if current_login_type == LoginType.USER_LOGIN:
            return current_user_client.user_uuid
        if current_login_type == LoginType.PARTICIPANT_LOGIN:
            return current_participant_client.participant_uuid
        if current_login_type == LoginType.DEVICE_LOGIN:
            return current_device_client.device_uuid
        if current_login_type == LoginType.SERVICE_LOGIN:
            return current_service_client.service_uuid

        return str()

    @defer.inlineCallbacks
    def register_to_events(self):
        # Always register to tests types events
        yield self.subscribe_pattern_with_callback(create_module_event_topic_from_name(
            ModuleNames.DATABASE_MODULE_NAME, 'test_type'), self.database_event_received)

        # Always register to tests events
        yield self.subscribe_pattern_with_callback(create_module_event_topic_from_name(
            ModuleNames.DATABASE_MODULE_NAME, 'test'), self.database_event_received)

        # Need to register to events (base class)
        super().register_to_events()

    def database_event_received(self, pattern, channel, message):
        # Process tests event
        try:
            tera_event = messages.TeraEvent()
            if isinstance(message, str):
                ret = tera_event.ParseFromString(message.encode('utf-8'))
            elif isinstance(message, bytes):
                ret = tera_event.ParseFromString(message)

            database_event = messages.DatabaseEvent()

            # Look for DatabaseEvent
            for any_msg in tera_event.events:
                if any_msg.Unpack(database_event):
                    self.handle_database_event(database_event)

        except DecodeError as decode_error:
            print('ServiceOpenTeraWithTests - DecodeError ', pattern, channel, message, decode_error)
        except ParseError as parse_error:
            print('ServiceOpenTeraWithTests - Failure in database_event_received', parse_error)

    def handle_database_event(self, event: messages.DatabaseEvent):
        if event.object_type == 'test':
            self.test_event_received(event)

        if event.object_type == 'test_type':
            self.test_type_event_received(event)

    @abstractmethod
    def test_event_received(self, event: messages.DatabaseEvent):
        pass

    @abstractmethod
    def test_type_event_received(self, event: messages.DatabaseEvent):
        pass



from opentera.forms.TeraForm import *

from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from flask_babel import gettext


class TeraServiceForm:

    @staticmethod
    def get_service_form():
        form = TeraForm("service")

        # Building lists
        #################
        # None to build!

        # Sections
        section = TeraFormSection("infos", gettext("Information"))
        form.add_section(section)

        # service_endpoint = db.Column(db.String, nullable=False)
        # service_clientendpoint = db.Column(db.String, nullable=False)
        # service_enabled = db.Column(db.Boolean, nullable=False, default=False)
        # Items
        section.add_item(TeraFormItem("id_service", gettext("Service ID"), "hidden", item_required=True))
        section.add_item(TeraFormItem("service_uuid", gettext("Service UUID"), "label", item_required=False,
                                      item_options={"readonly": False}))
        section.add_item(TeraFormItem("service_name", gettext("Service Name"), "text", item_required=True))
        section.add_item(TeraFormItem("service_key", gettext("Service Key"), "text", item_required=True))
        section.add_item(TeraFormItem("service_hostname", gettext("Service Hostname"), "text", item_required=True))
        section.add_item(TeraFormItem("service_port", gettext("Service Internal Port"), "numeric", item_required=True))
        section.add_item(TeraFormItem("service_endpoint", gettext("Service Internal URL"), "text", item_required=True))
        section.add_item(TeraFormItem("service_clientendpoint", gettext("Service External Endpoint"), "text",
                                      item_required=True))
        section.add_item(TeraFormItem("service_endpoint_user", gettext("Service User Endpoint"), "text",
                                      item_required=False))
        section.add_item(TeraFormItem("service_endpoint_participant", gettext("Service Participant Endpoint"), "text",
                                      item_required=False))
        section.add_item(TeraFormItem("service_endpoint_device", gettext("Service Device Endpoint"), "text",
                                      item_required=False))
        section.add_item(TeraFormItem("service_system", gettext("System Service"), "boolean", item_required=True))
        section.add_item(TeraFormItem("service_enabled", gettext("Service Enabled"), "boolean", item_required=True))
        section.add_item(TeraFormItem("service_editable_config", gettext("Service has editable config"), "boolean",
                                      item_required=False))
        section.add_item(TeraFormItem("service_default_config", gettext("Service Default Configuration"), "longtext",
                                      item_required=False))

        return form.to_dict()



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship


class TeraServiceRole(BaseModel, SoftDeleteMixin):
    __tablename__ = 't_services_roles'
    id_service_role = Column(Integer, Sequence('id_service_role_sequence'), primary_key=True, autoincrement=True)
    id_service = Column(Integer, ForeignKey('t_services.id_service', ondelete='cascade'), nullable=False)
    # Specific project role for a project, used mostly with OpenTera service for project access
    id_project = Column(Integer, ForeignKey('t_projects.id_project', ondelete='cascade'), nullable=True)
    # Specific site role for a site, used mostly with OpenTera service for site access
    id_site = Column(Integer, ForeignKey('t_sites.id_site', ondelete='cascade'), nullable=True)
    service_role_name = Column(String(100), nullable=False)

    service_role_service = relationship("TeraService", viewonly=True)
    service_role_project = relationship('TeraProject', viewonly=True)
    service_role_site = relationship('TeraSite', viewonly=True)

    service_role_user_groups = relationship("TeraUserGroup", secondary="t_services_access",
                                            back_populates="user_group_services_roles",  lazy='selectin')

    def __init__(self):
        pass

    def __str__(self):
        return '<TeraServiceRole Service = ' + str(self.service_role_service.service_name) + ', Role = ' + \
               str(self.service_role_name) + ' >'

    def __repr__(self):
        return self.__str__()

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['service_role_service', 'service_role_project', 'service_role_site',
                              'service_role_user_group'])

        if minimal:
            ignore_fields.extend([])

        json_val = super().to_json(ignore_fields=ignore_fields)

        # Remove null values
        if not json_val['id_project']:
            del json_val['id_project']
        else:
            if not minimal:
                json_val['project_name'] = self.service_role_project.project_name

        if not json_val['id_site']:
            del json_val['id_site']
        else:
            if not minimal:
                json_val['site_name'] = self.service_role_site.site_name

        if not minimal:
            json_val['service_name'] = self.service_role_service.service_name
        return json_val

    @staticmethod
    def get_service_roles(service_id: int, globals_only: bool = False, with_deleted: bool = False):
        query = TeraServiceRole.query.execution_options(include_deleted=with_deleted).filter_by(id_service=service_id)

        if globals_only:
            query = query.filter_by(id_site=None).filter_by(id_project=None)

        return query.all()

    @staticmethod
    def get_service_roles_for_site(service_id: int, site_id: int, with_deleted: bool = False):
        return TeraServiceRole.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=service_id, id_site=site_id).all()

    @staticmethod
    def get_service_roles_for_project(service_id: int, project_id: int, with_deleted: bool = False):
        return TeraServiceRole.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=service_id, id_project=project_id).all()

    @staticmethod
    def get_service_role_by_id(role_id: int, with_deleted: bool = False):
        return TeraServiceRole.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service_role=role_id).first()

    @staticmethod
    def get_service_role_by_name(service_id: int, rolename: str, site_id: int | None = None,
                                 project_id: int | None = None, with_deleted: bool = False):

        query = TeraServiceRole.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_service=service_id, service_role_name=rolename)

        if site_id:
            query = query.filter_by(id_site=site_id)

        if project_id:
            query = query.filter_by(id_project=project_id)

        return query.first()

    @staticmethod
    def create_defaults(test=False):
        from opentera.db.models.TeraService import TeraService

        for service in TeraService.query.all():
            if service.service_key != 'OpenTeraServer':  # Don't add global roles for TeraServer
                new_role = TeraServiceRole()
                new_role.id_service = service.id_service
                new_role.service_role_name = 'admin'
                TeraServiceRole.db().session.add(new_role)

                new_role = TeraServiceRole()
                new_role.id_service = service.id_service
                new_role.service_role_name = 'user'
                TeraServiceRole.db().session.add(new_role)
            else:
                pass  # TODO: do what we did in Project and Site Access

        TeraServiceRole.db().session.commit()



from opentera.redis.RedisProtocolFactory import RedisProtocolFactory, redisProtocol

# Twisted
from twisted.application import internet, service
from twisted.internet import reactor, ssl, defer

# Event based, twisted redis
import txredisapi as txredis

# Blocking redis
import redis


class RedisClient:

    def __init__(self, config=None):
        print('Init RedisClient', self, config)
        self.protocol = None
        self.callbacks_dict = dict()

        # Fill config
        if config is None:
            print('RedisClient - Warning, using default redis configuration')
            config = {'hostname': '127.0.0.1', 'port': 6379, 'db': 0, 'username': '', 'password': ''}

        self.redisConfig = config

        # Redis client (synchronous)
        self.redis = redis.Redis(host=config['hostname'],
                                 port=config['port'],
                                 db=config['db'],
                                 username=config['username'],
                                 password=config['password'],
                                 client_name=self.__class__.__name__)

        print('++++++++++++************** Connexion pool: ', self.redis.connection_pool)

        # Redis client (async)
        self.conn = reactor.connectTCP(config['hostname'], config['port'],
                                       RedisProtocolFactory(parent=self, protocol=redisProtocol, config=config))
        # print(self.conn)

    # def __del__(self):
    #     print("****- Deleting RedisClient")

    def getConfig(self):
        return self.redisConfig

    def redisConnectionMade(self):
        print(self, '********************* RedisClient connectionMade')
        pass

    def redisMessageReceived(self, pattern, channel, message):
        # print(self, 'redisMessageReceived', pattern, channel, message)
        if pattern in self.callbacks_dict:
            # Call function.
            self.callbacks_dict[pattern](pattern, channel, message)

    def redisConnectionLost(self, reason):
        print(self, "RedisClient lost connection", reason)
        pass

    def setProtocol(self, protocol: redisProtocol):
        print('RedisClient set protocol', protocol)
        self.protocol = protocol

    def subscribe(self, topic):
        if self.protocol:
            # print('RedisClient (', self, ') subscribing to: ',  topic)
            return self.protocol.psubscribe(topic)
        else:
            print('Error, no protocol')
            return defer.Deferred()

    def unsubscribe(self, topic):
        if self.protocol:
            return self.protocol.punsubscribe(topic)
        else:
            print('Error, no protocol')
            return defer.Deferred()

    def publish(self, topic, message):
        return self.redis.publish(topic, message)

    def redisGet(self, key):
        return self.redis.get(key)

    def redisSet(self, key, value, ex=None):
        return self.redis.set(key, value, ex=ex)

    def redisDelete(self, key):
        return self.redis.delete(key)

    def subscribe_pattern_with_callback(self, pattern, function):
        # print(self, 'subscribe_pattern_with_callback', pattern, function)
        self.callbacks_dict[pattern] = function
        return self.subscribe(pattern)

    def unsubscribe_pattern_with_callback(self, pattern, function):
        # print(self, 'unsubscribe_pattern_with_callback', pattern, function)
        ret = self.unsubscribe(pattern)
        del self.callbacks_dict[pattern]
        return ret

    @defer.inlineCallbacks
    def redisClose(self):
        if self.protocol:
            # Will not reconnect
            var = yield self.protocol.quit()
            self.protocol.parent = None
            self.protocol.factory.parent = None
            self.protocol = None
            print(var)

        if self.conn:
            # Disconnect socket
            self.conn.disconnect()

        if self.redis:
            # Close sync client (will terminate conn.)
            self.redis.close()


# Debug
if __name__ == '__main__':
    print('Starting')

    from twisted.internet.task import deferLater
    from twisted.internet.defer import inlineCallbacks


    def sleep(secs):
        return deferLater(reactor, secs, lambda: None)


    @inlineCallbacks
    def function_with_client_should_remove_conn():
        class MyClient(RedisClient):
            def redisConnectionMade(self):
                print('redisConnectionMade')
                self.subscribe('*')

        client = MyClient()
        print('setting variable')
        client.redisSet('papa', 'rien', ex=60)
        print('redis get', client.redisGet('papa'))
        print('sleeping 2 secs.')
        yield sleep(2)
        client.redisClose()

    def called(result, msg: str = None):
        print('Function called', msg)

    from twisted.internet import task
    d1 = task.deferLater(reactor, 1, function_with_client_should_remove_conn)
    d1.addCallback(called, 'function_with_client_should_remove_conn done!')

    # RPCClient
    from opentera.redis.RedisRPCClient import RedisRPCClient
    from opentera.config.ConfigManager import ConfigManager

    def function_with_rpc_client_should_remove_conn():
        config = ConfigManager()
        config.create_defaults()

        rpc = RedisRPCClient(config.redis_config, timeout=1)
        rpc.call('module', 'func', 1, 2, 3)


    d2 = task.deferLater(reactor, 1, function_with_rpc_client_should_remove_conn)
    d2.addCallback(called, 'function_with_rpc_client_should_remove_conn done!')


    def collect():
        print('collecting...')
        import gc
        gc.collect()

    d3 = task.LoopingCall(collect)
    d3.start(5)

    print('Starting reactor')
    reactor.run()






import jwt
import json
import time
from opentera.redis.RedisVars import RedisVars
from opentera.redis.RedisClient import RedisClient
from requests import get, post, Response, delete
from opentera.services.ServiceConfigManager import ServiceConfigManager
import opentera.messages.python as messages
from twisted.internet import defer
import datetime
from opentera.logging.LoggingClient import LoggingClient
from opentera.services.ServiceAccessManager import ServiceAccessManager


class ServiceOpenTera(RedisClient):

    def __init__(self, config_man: ServiceConfigManager, service_info):
        # First initialize redis
        RedisClient.__init__(self, config_man.redis_config)

        # Initialize logger
        self.logger = LoggingClient(config_man.redis_config, 'LoggingClient_' + self.__class__.__name__)

        # Store service info
        self.service_info = service_info

        # Store RPC API
        self.rpc_api = dict()

        # Store config
        self.config = config_man.service_config
        self.config_man = config_man

        # Take values from config_man
        # Values are checked when config is loaded...
        self.backend_hostname = config_man.backend_config['hostname']
        self.backend_port = config_man.backend_config['port']
        self.service_uuid = config_man.service_config['ServiceUUID']

        # Create service token for service api requests
        self.service_token = self.service_generate_token()

        # Init service access manager
        ServiceAccessManager.init_access_manager(service=self)

    def redisConnectionMade(self):
        print('*** ServiceOpenTera.redisConnectionMade for', self.config['name'])

        # Build RPC interface
        self.setup_rpc_interface()

        # Build standard interface
        self.build_interface()

        # Register to system events
        self.register_to_events()

    def setup_rpc_interface(self):
        self.rpc_api['session_type_config'] = {'args': ['int:id_session_type'],
                                               'returns': 'dict',
                                               'callback': self.get_session_type_config_form}

    def register_to_events(self):
        # Should be implemented in derived classes
        yield None

    def notify_service_messages(self, pattern, channel, message):
        pass

    @defer.inlineCallbacks
    def build_interface(self):
        # TODO not sure of the interface using UUID or name here...
        # Will do  both!
        ret1 = yield self.subscribe_pattern_with_callback(
            RedisVars.build_service_message_topic( self.service_info['service_uuid']), self.notify_service_messages)

        ret2 = yield self.subscribe_pattern_with_callback(
            RedisVars.build_service_message_topic(self.service_info['service_key']), self.notify_service_messages)

        ret3 = yield self.subscribe_pattern_with_callback(
            RedisVars.build_service_rpc_topic(self.service_info['service_uuid']), self.notify_service_rpc)

        ret4 = yield self.subscribe_pattern_with_callback(
            RedisVars.build_service_rpc_topic(self.service_info['service_key']), self.notify_service_rpc)

        print(ret1, ret2, ret3, ret4)

    def notify_service_rpc(self, pattern, channel, message):
        import threading
        print('ServiceOpenTera - Received rpc', self, pattern, channel, message, ' thread:', threading.current_thread())

        rpc_message = messages.RPCMessage()

        try:
            # Look for a RPCMessage
            rpc_message.ParseFromString(message)

            if self.rpc_api.__contains__(rpc_message.method):

                # RPC method found, call it with the args
                args = list()
                kwargs = dict()

                # TODO type checking with declared rpc interface ?
                for value in rpc_message.args:
                    # Append the oneof value to args
                    args.append(getattr(value, value.WhichOneof('arg_value')))

                # Call callback function
                ret_value = self.rpc_api[rpc_message.method]['callback'](*args, **kwargs)

                # More than we need?
                my_dict = {'method': rpc_message.method,
                           'id': rpc_message.id,
                           'pattern': pattern,
                           'status': 'OK',
                           'return_value': ret_value}

                json_data = json.dumps(my_dict)

                # Return result (a json string)
                self.publish(rpc_message.reply_to, json_data)

        except Exception as e:
            import sys
            print('Error calling rpc method', message, sys.exc_info(), e)
            my_dict = {'method': rpc_message.method,
                       'id': rpc_message.id,
                       'pattern': pattern,
                       'status': 'Error',
                       'return_value': None}

            json_data = json.dumps(my_dict)

            # Return result (a json string)
            self.publish(rpc_message.reply_to, json_data)

    def service_generate_token(self):
        # Use redis key to generate token
        # Creating token with service info
        # TODO ADD MORE FIELDS?
        payload = {
            'iat': int(time.time()),
            'service_uuid': self.service_uuid
        }

        return jwt.encode(payload, self.redisGet(RedisVars.RedisVar_ServiceTokenAPIKey), algorithm='HS256')

    def post_to_opentera(self, api_url: str, json_data: dict) -> Response:
        # Synchronous call to OpenTera backend
        url = "https://" + self.backend_hostname + ':' + str(self.backend_port) + api_url
        request_headers = {'Authorization': 'OpenTera ' + self.service_token}
        return post(url=url, verify=False, headers=request_headers, json=json_data)

    def get_from_opentera(self, api_url: str, params: dict) -> Response:
        from flask import jsonify, Response
        # Synchronous call to OpenTera backend
        url = "https://" + self.backend_hostname + ':' + str(self.backend_port) + api_url
        request_headers = {'Authorization': 'OpenTera ' + self.service_token}
        return get(url=url, verify=False, headers=request_headers, params=params)

    def delete_from_opentera(self, api_url: str, params: dict) -> Response:
        # Synchronous call to OpenTera backend
        url = "https://" + self.backend_hostname + ':' + str(self.backend_port) + api_url
        request_headers = {'Authorization': 'OpenTera ' + self.service_token}
        return delete(url=url, verify=False, headers=request_headers, params=params)

    def send_event_message(self, event, topic: str):
        message = self.create_event_message(topic)
        any_message = messages.Any()
        any_message.Pack(event)
        message.events.extend([any_message])
        return self.publish(message.header.topic, message.SerializeToString())

    def create_event_message(self, topic):
        event_message = messages.TeraEvent()
        event_message.header.version = 1
        event_message.header.time = datetime.datetime.now().timestamp()
        event_message.header.topic = topic
        return event_message

    def send_tera_message(self, event, src: str, topic: str):
        message = self.create_tera_message(src, topic)
        any_message = messages.Any()
        any_message.Pack(event)
        message.data.extend([any_message])
        return self.publish(topic, message.SerializeToString())

    def create_tera_message(self, src='', dest='', seq=0):
        tera_message = messages.TeraModuleMessage()
        tera_message.head.version = 1
        tera_message.head.time = datetime.datetime.now().timestamp()
        tera_message.head.seq = seq
        tera_message.head.source = src
        tera_message.head.dest = dest
        return tera_message

    def get_session_type_config_form(self, id_session_type: int) -> dict:
        # Default session type config form for services
        return {}



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from opentera.db.models.TeraTestTypeProject import TeraTestTypeProject
from sqlalchemy import Column, ForeignKey, Integer, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraTestTypeSite(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_tests_types_sites'
    id_test_type_site = Column(Integer, Sequence('id_test_type_site_sequence'), primary_key=True,
                               autoincrement=True)
    id_test_type = Column('id_test_type', Integer, ForeignKey('t_tests_types.id_test_type', ondelete='cascade'),
                          nullable=False)
    id_site = Column('id_site', Integer, ForeignKey('t_sites.id_site', ondelete='cascade'), nullable=False)

    test_type_site_test_type = relationship("TeraTestType", viewonly=True)
    test_type_site_site = relationship("TeraSite", viewonly=True)

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.extend(['test_type_site_test_type', 'test_type_site_site'])

        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraTestType import TeraTestType
            from opentera.db.models.TeraSite import TeraSite

            pre_test = TeraTestType.get_test_type_by_id(1)
            post_test = TeraTestType.get_test_type_by_id(2)
            general_test = TeraTestType.get_test_type_by_id(3)

            default_site = TeraSite.get_site_by_sitename('Default Site')
            secret_site = TeraSite.get_site_by_sitename('Top Secret Site')

            tts = TeraTestTypeSite()
            tts.id_test_type = pre_test.id_test_type
            tts.id_site = default_site.id_site
            TeraTestTypeSite.db().session.add(tts)

            tts = TeraTestTypeSite()
            tts.id_test_type = post_test.id_test_type
            tts.id_site = default_site.id_site
            TeraTestTypeSite.db().session.add(tts)

            tts = TeraTestTypeSite()
            tts.id_test_type = pre_test.id_test_type
            tts.id_site = secret_site.id_site
            TeraTestTypeSite.db().session.add(tts)

            tts = TeraTestTypeSite()
            tts.id_test_type = general_test.id_test_type
            tts.id_site = secret_site.id_site
            TeraTestTypeSite.db().session.add(tts)

            TeraTestTypeSite.db().session.commit()

    @staticmethod
    def get_test_type_site_by_id(tts_id: int, with_deleted: bool = False):
        return TeraTestTypeSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_test_type_site=tts_id).first()

    @staticmethod
    def get_sites_for_test_type(test_type_id: int, with_deleted: bool = False):
        return TeraTestTypeSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_test_type=test_type_id).all()

    @staticmethod
    def get_tests_types_for_site(site_id: int, with_deleted: bool = False):
        return TeraTestTypeSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_site=site_id).all()

    @staticmethod
    def get_test_type_site_for_test_type_and_site(site_id: int, test_type_id: int, with_deleted: bool = False):
        return TeraTestTypeSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_site=site_id, id_test_type=test_type_id).first()

    @staticmethod
    def get_test_type_site_for_site_and_service(site_id: int, service_id: int, with_deleted: bool = False):
        from opentera.db.models.TeraTestType import TeraTestType
        return TeraTestTypeSite.query.execution_options(include_deleted=with_deleted).join(TeraTestType). \
            filter(TeraTestType.id_service == service_id). \
            filter(TeraTestTypeSite.id_site == site_id).all()

    @staticmethod
    def check_integrity(obj_to_check):
        # If that test type is related to a service, make sure that the service is associated to that site
        service_sites = [site.id_site for site in
                         obj_to_check.test_type_site_test_type.test_type_service.service_sites]
        if obj_to_check.id_site not in service_sites:
            # We must also associate that service to that site!
            from opentera.db.models.TeraServiceSite import TeraServiceSite
            new_service_site = TeraServiceSite()
            new_service_site.id_service = obj_to_check.test_type_site_test_type.test_type_service.id_service
            new_service_site.id_site = obj_to_check.id_site
            TeraServiceSite.insert(new_service_site)

    @staticmethod
    def delete_with_ids(test_type_id: int, site_id: int, autocommit: bool = True):
        delete_obj: TeraTestTypeSite = TeraTestTypeSite.query.filter_by(id_test_type=test_type_id,
                                                                        id_site=site_id).first()
        if delete_obj:
            TeraTestTypeSite.delete(delete_obj.id_test_type_site, autocommit=autocommit)

    @classmethod
    def delete(cls, id_todel, autocommit: bool = True):
        from opentera.db.models.TeraTestTypeProject import TeraTestTypeProject
        # Delete all association with projects for that site
        delete_obj = TeraTestTypeSite.query.filter_by(id_test_type_site=id_todel).first()

        if delete_obj:
            projects = TeraTestTypeProject.get_projects_for_test_type(delete_obj.id_test_type)
            for tt_project in projects:
                if tt_project.test_type_project_project.id_site == delete_obj.id_site:
                    TeraTestTypeProject.delete(tt_project.id_test_type_project, autocommit=autocommit)

            # Ok, delete it
            super().delete(id_todel, autocommit=autocommit)

    @classmethod
    def insert(cls, tts):
        inserted_obj = super().insert(tts)
        TeraTestTypeSite.check_integrity(inserted_obj)
        return inserted_obj

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        for project in self.test_type_site_site.site_projects:
            test_type_project = TeraTestTypeProject.get_test_type_project_for_test_type_project(project.id_project,
                                                                                                self.id_test_type,
                                                                                                with_deleted)
            if test_type_project:
                cannot_be_deleted_exception = test_type_project.delete_check_integrity(with_deleted=with_deleted)
                if cannot_be_deleted_exception:
                    return IntegrityError('Still have test of that type in the site', self.id_test_type, 't_tests')
        return None

    @classmethod
    def update(cls, update_id: int, values: dict):
        return



from opentera.forms.TeraForm import *
from flask_babel import gettext


class TeraSiteForm:

    @staticmethod
    def get_site_form():
        form = TeraForm("site")

        # Sections
        section = TeraFormSection("informations", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_site", gettext("Site ID"), "hidden", True))
        section.add_item(TeraFormItem("site_name", gettext("Site Name"), "text", True))
        section.add_item(TeraFormItem("site_role", gettext("Site Role"), "hidden", False))

        return form.to_dict()



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship


class TeraSessionUsers(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_sessions_users'
    id_session_user = Column(Integer, Sequence('id_session_user'), primary_key=True, autoincrement=True)
    id_session = Column(Integer, ForeignKey('t_sessions.id_session', ondelete='cascade'))
    id_user = Column(Integer, ForeignKey('t_users.id_user'))

    session_user_session = relationship('TeraSession', viewonly=True)
    session_user_user = relationship('TeraUser', viewonly=True)

    @staticmethod
    def get_session_count_for_user(id_user: int, with_deleted: bool = False) -> int:
        return TeraSessionUsers.get_count(filters={'id_user': id_user}, with_deleted=with_deleted)

    @classmethod
    def update(cls, update_id: int, values: dict):
        return



import uuid

from requests import Response
from flask import request


class TeraDeviceClient:

    def __init__(self, token_dict: dict, token: str, config_man):
        self.__device_uuid = token_dict['device_uuid']
        self.__id_device = token_dict['id_device']
        self.__device_token = token

        # A little trick here to get the right URL for the server if we are using a proxy
        backend_hostname = config_man.backend_config["hostname"]
        backend_port = str(config_man.backend_config["port"])

        # if 'X-Externalhost' in request.headers:
        #    backend_hostname = request.headers['X-Externalhost']

        #if 'X-Externalport' in request.headers:
        #    backend_port = request.headers['X-Externalport']

        self.__backend_url = 'https://' + backend_hostname + ':' + backend_port


    @property
    def device_uuid(self):
        return self.__device_uuid

    @device_uuid.setter
    def device_uuid(self, u_uuid: uuid):
        self.__device_uuid = u_uuid

    @property
    def id_device(self):
        return self.__id_device

    @id_device.setter
    def id_device(self, id_dev: int):
        self.__id_device = id_dev

    @property
    def device_token(self):
        return self.__device_token

    @device_token.setter
    def device_token(self, token: str):
        self.__device_token = token

    def do_get_request_to_backend(self, path: str) -> Response:
        from requests import get
        request_headers = {'Authorization': 'OpenTera ' + self.__device_token}
        # TODO: remove verify=False and check certificate
        backend_response = get(url=self.__backend_url + path, headers=request_headers, verify=False)
        return backend_response

    def can_access_session(self, id_session: int) -> bool:
        response = self.do_get_request_to_backend('/api/device/sessions?id_session=' + str(id_session))
        return response.status_code == 200

    def get_device_infos(self) -> dict:
        response = self.do_get_request_to_backend('/api/device/devices')
        if response.status_code == 200:
            return response.json()
        return {}

    def __repr__(self):
        return '<TeraDeviceClient - UUID: ' + self.__device_token \
               + ', Token: ' + self.__device_token + '>'



from opentera.db.Base import BaseModel
from opentera.db.models.TeraSessionTypeProject import TeraSessionTypeProject
from opentera.db.models.TeraSessionType import TeraSessionType
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraSessionTypeSite(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_sessions_types_sites'
    id_session_type_site = Column(Integer, Sequence('id_session_type_site_sequence'), primary_key=True,
                                  autoincrement=True)
    id_session_type = Column('id_session_type', Integer, ForeignKey('t_sessions_types.id_session_type',
                                                                    ondelete='cascade'), nullable=False)
    id_site = Column('id_site', Integer, ForeignKey('t_sites.id_site', ondelete='cascade'), nullable=False)

    session_type_site_session_type = relationship("TeraSessionType", viewonly=True)
    session_type_site_site = relationship("TeraSite", viewonly=True)

    def to_json(self, ignore_fields=[], minimal=False):
        ignore_fields.extend(['session_type_site_session_type', 'session_type_site_site'])

        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    def to_json_create_event(self):
        json_data = self.to_json(minimal=True)
        # Query session type information
        session_type = TeraSessionType.get_session_type_by_id(self.id_session_type)
        if session_type:
            json_data['session_type_id_service'] = session_type.id_service
        return json_data

    def to_json_update_event(self):
        json_data = self.to_json(minimal=True)
        json_data['session_type_id_service'] = self.session_type_site_session_type.id_service
        return json_data

    def to_json_delete_event(self):
        # Minimal information, delete can not be filtered
        return {'id_session_type_site': self.id_session_type_site}

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraSessionType import TeraSessionType
            from opentera.db.models.TeraSite import TeraSite

            video_session = TeraSessionType.get_session_type_by_id(1)
            sensor_session = TeraSessionType.get_session_type_by_id(2)
            data_session = TeraSessionType.get_session_type_by_id(3)
            exerc_session = TeraSessionType.get_session_type_by_id(4)
            bureau_session = TeraSessionType.get_session_type_by_id(5)

            default_site = TeraSite.get_site_by_sitename('Default Site')
            secret_site = TeraSite.get_site_by_sitename('Top Secret Site')

            sts = TeraSessionTypeSite()
            sts.id_session_type = video_session.id_session_type
            sts.id_site = default_site.id_site
            TeraSessionTypeSite.db().session.add(sts)

            sts = TeraSessionTypeSite()
            sts.id_session_type = sensor_session.id_session_type
            sts.id_site = default_site.id_site
            TeraSessionTypeSite.db().session.add(sts)

            sts = TeraSessionTypeSite()
            sts.id_session_type = data_session.id_session_type
            sts.id_site = default_site.id_site
            TeraSessionTypeSite.db().session.add(sts)

            sts = TeraSessionTypeSite()
            sts.id_session_type = exerc_session.id_session_type
            sts.id_site = default_site.id_site
            TeraSessionTypeSite.db().session.add(sts)

            sts = TeraSessionTypeSite()
            sts.id_session_type = bureau_session.id_session_type
            sts.id_site = default_site.id_site
            TeraSessionTypeSite.db().session.add(sts)

            sts = TeraSessionTypeSite()
            sts.id_session_type = exerc_session.id_session_type
            sts.id_site = secret_site.id_site
            TeraSessionTypeSite.db().session.add(sts)

            TeraSessionTypeSite.db().session.commit()
        else:
            # Automatically associate session types that are in a project to that site
            from opentera.db.models.TeraSessionTypeProject import TeraSessionTypeProject
            for stp in TeraSessionTypeProject.query_with_filters():
                project_site_id = stp.session_type_project_project.id_site
                if not TeraSessionTypeSite.get_session_type_site_for_session_type_and_site(site_id=project_site_id,
                                                                                           session_type_id=
                                                                                           stp.id_session_type):
                    # No association - create a new one
                    st_site = TeraSessionTypeSite()
                    st_site.id_site = project_site_id
                    st_site.id_session_type = stp.id_session_type
                    TeraSessionTypeSite.db().session.add(st_site)
                    TeraSessionTypeSite.db().session.commit()

    @staticmethod
    def get_session_type_site_by_id(sts_id: int, with_deleted: bool = False):
        return TeraSessionTypeSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_session_type_site=sts_id).first()

    @staticmethod
    def get_sites_for_session_type(session_type_id: int, with_deleted: bool = False):
        return TeraSessionTypeSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_session_type=session_type_id).all()

    @staticmethod
    def get_sessions_types_for_site(site_id: int, with_deleted: bool = False):
        return TeraSessionTypeSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_site=site_id).all()

    @staticmethod
    def get_session_type_site_for_session_type_and_site(site_id: int, session_type_id: int, with_deleted: bool = False):
        return TeraSessionTypeSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_site=site_id, id_session_type=session_type_id).first()

    @staticmethod
    def get_session_type_site_for_site_and_service(site_id: int, service_id: int, with_deleted: bool = False):
        from opentera.db.models.TeraSessionType import TeraSessionType
        return TeraSessionTypeSite.query.execution_options(include_deleted=with_deleted).join(TeraSessionType). \
            filter(TeraSessionType.id_service == service_id). \
            filter(TeraSessionTypeSite.id_site == site_id).all()

    @staticmethod
    def check_integrity(obj_to_check):
        from opentera.db.models.TeraSessionType import TeraSessionType
        # If that session type is related to a service, make sure that the service is associated to that site
        if obj_to_check.session_type_site_session_type.session_type_category == \
                TeraSessionType.SessionCategoryEnum.SERVICE.value:
            service_sites = [site.id_site for site in
                             obj_to_check.session_type_site_session_type.session_type_service.service_sites]
            if obj_to_check.id_site not in service_sites:
                # We must also associate that service to that site!
                from opentera.db.models.TeraServiceSite import TeraServiceSite
                new_service_site = TeraServiceSite()
                new_service_site.id_service = obj_to_check.session_type_site_session_type \
                    .session_type_service.id_service
                new_service_site.id_site = obj_to_check.id_site
                TeraServiceSite.insert(new_service_site)

    @staticmethod
    def delete_with_ids(session_type_id: int, site_id: int, autocommit: bool = True):
        delete_obj: TeraSessionTypeSite = TeraSessionTypeSite.query.filter_by(id_session_type=session_type_id,
                                                                              id_site=site_id).first()
        if delete_obj:
            TeraSessionTypeSite.delete(delete_obj.id_session_type_site, autocommit=autocommit)

    @classmethod
    def delete(cls, id_todel, autocommit: bool = True):
        from opentera.db.models.TeraSessionTypeProject import TeraSessionTypeProject
        # Delete all association with projects for that site
        delete_obj = TeraSessionTypeSite.query.filter_by(id_session_type_site=id_todel).first()

        if delete_obj:
            projects = TeraSessionTypeProject.get_projects_for_session_type(delete_obj.id_session_type)
            for st_project in projects:
                if st_project.session_type_project_project.id_site == delete_obj.id_site:
                    TeraSessionTypeProject.delete(st_project.id_session_type_project, autocommit=autocommit)

            # Ok, delete it
            super().delete(id_todel, autocommit=autocommit)

    @classmethod
    def insert(cls, sts):
        inserted_obj = super().insert(sts)
        TeraSessionTypeSite.check_integrity(inserted_obj)
        return inserted_obj

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        for project in self.session_type_site_site.site_projects:
            ses_type_project = TeraSessionTypeProject.get_session_type_project_for_session_type_project(
                project.id_project, self.id_session_type, with_deleted=with_deleted)
            if ses_type_project:
                cannot_be_deleted_exception = ses_type_project.delete_check_integrity(with_deleted=with_deleted)
                if cannot_be_deleted_exception:
                    return IntegrityError('Still have sessions of that type in the site', self.id_session_type,
                                          't_sessions')
        return None

    @classmethod
    def update(cls, update_id: int, values: dict):
        return



from .TeraAsset import TeraAsset
from .TeraDevice import TeraDevice
from .TeraDeviceParticipant import TeraDeviceParticipant
from .TeraDeviceProject import TeraDeviceProject
from .TeraDeviceSite import TeraDeviceSite
from .TeraDeviceSubType import TeraDeviceSubType
from .TeraDeviceType import TeraDeviceType
from .TeraParticipant import TeraParticipant
from .TeraParticipantGroup import TeraParticipantGroup
from .TeraProject import TeraProject
from .TeraServerSettings import TeraServerSettings
from .TeraService import TeraService
from .TeraServiceAccess import TeraServiceAccess
from .TeraServiceConfig import TeraServiceConfig
from .TeraServiceConfigSpecific import TeraServiceConfigSpecific
from .TeraServiceProject import TeraServiceProject
from .TeraServiceRole import TeraServiceRole
from .TeraServiceSite import TeraServiceSite
from .TeraSession import TeraSession
from .TeraSessionDevices import TeraSessionDevices
from .TeraSessionEvent import TeraSessionEvent
from .TeraSessionParticipants import TeraSessionParticipants
from .TeraSessionType import TeraSessionType
from .TeraSessionTypeProject import TeraSessionTypeProject
from .TeraSessionTypeSite import TeraSessionTypeSite
from .TeraSessionUsers import TeraSessionUsers
from .TeraSite import TeraSite
from .TeraTest import TeraTest
from .TeraTestType import TeraTestType
from .TeraTestTypeProject import TeraTestTypeProject
from .TeraTestTypeSite import TeraTestTypeSite
from .TeraUser import TeraUser
from .TeraUserGroup import TeraUserGroup
from .TeraUserUserGroup import TeraUserUserGroup
from .TeraUserPreference import TeraUserPreference
from .TeraUserUserGroup import TeraUserUserGroup

"""
    A map containing the event name and class, useful for event filtering. 
    Insert only useful events here.
"""
EventNameClassMap = {
    TeraAsset.get_model_name(): TeraAsset,
    TeraDevice.get_model_name(): TeraDevice,
    TeraParticipant.get_model_name(): TeraParticipant,
    TeraParticipantGroup.get_model_name(): TeraParticipantGroup,
    TeraProject.get_model_name(): TeraProject,
    TeraSession.get_model_name(): TeraSession,
    TeraSessionType.get_model_name(): TeraSessionType,
    TeraSite.get_model_name(): TeraSite,
    TeraUser.get_model_name(): TeraUser,
    TeraUserGroup.get_model_name(): TeraUserGroup,
    TeraTestType.get_model_name(): TeraTestType,
    TeraTest.get_model_name(): TeraTest,
    TeraService.get_model_name(): TeraService,
    TeraSessionTypeSite.get_model_name(): TeraSessionTypeSite,
    TeraSessionTypeProject.get_model_name(): TeraSessionTypeProject,
    TeraServerSettings.get_model_name(): TeraServerSettings
}

# All exported symbols
__all__ = ['TeraAsset',
           'TeraDevice',
           'TeraDeviceParticipant',
           'TeraDeviceProject',
           'TeraDeviceSite',
           'TeraDeviceSubType',
           'TeraDeviceType',
           'TeraParticipant',
           'TeraParticipantGroup',
           'TeraProject',
           'TeraServerSettings',
           'TeraService',
           'TeraServiceAccess',
           'TeraServiceConfig',
           'TeraServiceConfigSpecific',
           'TeraServiceProject',
           'TeraServiceRole',
           'TeraServiceSite',
           'TeraSession',
           'TeraSessionDevices',
           'TeraSessionEvent',
           'TeraSessionParticipants',
           'TeraSessionType',
           'TeraSessionTypeProject',
           'TeraSessionTypeSite',
           'TeraSessionUsers',
           'TeraSite',
           'TeraTest',
           'TeraTestType',
           'TeraTestTypeSite',
           'TeraTestTypeProject',
           'TeraUser',
           'TeraUserGroup',
           'TeraUserPreference',
           'TeraUserUserGroup',
           'EventNameClassMap'
           ]



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraDeviceSite(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_devices_sites'
    id_device_site = Column(Integer, Sequence('id_device_site_sequence'), primary_key=True, autoincrement=True)
    id_device = Column(Integer, ForeignKey("t_devices.id_device", ondelete='cascade'), nullable=False)
    id_site = Column(Integer, ForeignKey("t_sites.id_site", ondelete='cascade'), nullable=False)

    device_site_site = relationship("TeraSite", viewonly=True)
    device_site_device = relationship("TeraDevice", viewonly=True)

    def to_json(self, ignore_fields=[], minimal=False):
        ignore_fields.extend(['device_site_site', 'device_site_device'])

        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraDevice import TeraDevice
            from opentera.db.models.TeraSite import TeraSite

            device1 = TeraDevice.get_device_by_name('Apple Watch #W05P1')
            device2 = TeraDevice.get_device_by_name('Kit Télé #1')
            # device3 = TeraDevice.get_device_by_name('Robot A')

            default_site = TeraSite.get_site_by_sitename('Default Site')
            secret_site = TeraSite.get_site_by_sitename('Top Secret Site')

            dev_site = TeraDeviceSite()
            dev_site.id_device = device1.id_device
            dev_site.id_site = default_site.id_site
            TeraDeviceSite.db().session.add(dev_site)

            dev_site = TeraDeviceSite()
            dev_site.id_device = device2.id_device
            dev_site.id_site = default_site.id_site
            TeraDeviceSite.db().session.add(dev_site)

            dev_site = TeraDeviceSite()
            dev_site.id_device = device1.id_device
            dev_site.id_site = secret_site.id_site
            TeraDeviceSite.db().session.add(dev_site)

            TeraDeviceSite.db().session.commit()
        else:
            # Automatically associate devices that are in a project to that site
            from opentera.db.models.TeraDeviceProject import TeraDeviceProject
            for dp in TeraDeviceProject.query_with_filters():
                project_site_id = dp.device_project_project.id_site
                if not TeraDeviceSite.get_device_site_id_for_device_and_site(site_id=project_site_id,
                                                                             device_id=
                                                                             dp.device_project_device.id_device):
                    # No association - create a new one
                    device_site = TeraDeviceSite()
                    device_site.id_site = project_site_id
                    device_site.id_device = dp.device_project_device.id_device
                    TeraDeviceSite.db().session.add(device_site)
                    TeraDeviceSite.db().session.commit()

    @staticmethod
    def get_device_site_by_id(device_site_id: int, with_deleted: bool = False):
        return TeraDeviceSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_device_site=device_site_id).first()

    @staticmethod
    def get_device_site_id_for_device_and_site(device_id: int, site_id: int, with_deleted: bool = False):
        return TeraDeviceSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_site=site_id, id_device=device_id).first()

    @staticmethod
    def get_devices_for_site(site_id: int, with_deleted: bool = False):
        return TeraDeviceSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_site=site_id).all()

    @staticmethod
    def get_sites_for_device(device_id: int, with_deleted: bool = False):
        return TeraDeviceSite.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_device=device_id).all()

    @staticmethod
    def delete_with_ids(device_id: int, site_id: int, autocommit: bool = True):
        delete_obj: TeraDeviceSite = TeraDeviceSite.query.filter_by(id_device=device_id, id_site=site_id).first()
        if delete_obj:
            TeraDeviceSite.delete(delete_obj.id_device_site, autocommit=autocommit)

    @classmethod
    def delete(cls, id_todel, autocommit: bool = True):
        from opentera.db.models.TeraDeviceProject import TeraDeviceProject

        delete_obj = TeraDeviceSite.query.filter_by(id_device_site=id_todel).first()
        if delete_obj:
            # Needed for device-project deletion later
            id_site = delete_obj.id_site
            id_device = delete_obj.id_device
            # Ok, delete it (will check integrity)
            super().delete(id_todel, autocommit)

            # Delete all association with projects for that site
            from opentera.db.models.TeraProject import TeraProject
            specific_device_projects = TeraDeviceProject.query.join(TeraProject). \
                filter(TeraDeviceProject.id_device == id_device).\
                filter(TeraProject.id_site == id_site).all()

            for device_project in specific_device_projects:
                TeraDeviceProject.delete(device_project.id_device_project, autocommit=autocommit)

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        from opentera.db.models.TeraDeviceProject import TeraDeviceProject
        from opentera.db.models.TeraProject import TeraProject

        # Will check if device is part of a project in the site
        specific_device_projects = TeraDeviceProject.query.join(TeraProject).\
            filter(TeraDeviceProject.id_device == self.id_device).filter(TeraProject.id_site == self.id_site).\
            execution_options(include_deleted=with_deleted).all()

        # Check integrity of device_projects
        for device_project in specific_device_projects:
            integrity_check = device_project.delete_check_integrity(with_deleted)
            if integrity_check is not None:
                return integrity_check

        # Everything is good
        return None

    @classmethod
    def update(cls, update_id: int, values: dict):
        return



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraDeviceParticipant(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_devices_participants'
    id_device_participant = Column(Integer, Sequence('id_device_participant_sequence'), primary_key=True,
                                   autoincrement=True)
    id_device = Column(Integer, ForeignKey("t_devices.id_device"), nullable=False)
    id_participant = Column(Integer, ForeignKey("t_participants.id_participant", ondelete='cascade'),
                            nullable=False)

    device_participant_participant = relationship("TeraParticipant", viewonly=True)
    device_participant_device = relationship("TeraDevice", viewonly=True)

    def to_json(self, ignore_fields=[], minimal=False):
        ignore_fields.extend(['device_participant_participant', 'device_participant_device'])

        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraParticipant import TeraParticipant
            from opentera.db.models.TeraDevice import TeraDevice
            participant1 = TeraParticipant.get_participant_by_id(1)
            participant2 = TeraParticipant.get_participant_by_id(2)
            device1 = TeraDevice.get_device_by_name('Apple Watch #W05P1')
            device2 = TeraDevice.get_device_by_name('Kit Télé #1')
            device3 = TeraDevice.get_device_by_name('Robot A')

            dev_participant = TeraDeviceParticipant()
            dev_participant.id_device = device1.id_device
            dev_participant.id_participant = participant1.id_participant
            TeraDeviceParticipant.db().session.add(dev_participant)

            dev_participant = TeraDeviceParticipant()
            dev_participant.id_device = device1.id_device
            dev_participant.id_participant = participant2.id_participant
            TeraDeviceParticipant.db().session.add(dev_participant)

            dev_participant = TeraDeviceParticipant()
            dev_participant.id_device = device2.id_device
            dev_participant.id_participant = participant2.id_participant
            TeraDeviceParticipant.db().session.add(dev_participant)

            TeraDeviceParticipant.db().session.commit()

    @staticmethod
    def get_device_participant_by_id(device_participant_id: int, with_deleted: bool = False):
        return TeraDeviceParticipant.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_device_participant=device_participant_id).first()

    @staticmethod
    def query_devices_for_participant(participant_id: int, with_deleted: bool = False):
        return TeraDeviceParticipant.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_participant=participant_id).all()

    @staticmethod
    def query_participants_for_device(device_id: int, with_deleted: bool = False):
        return TeraDeviceParticipant.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_device=device_id).all()

    @staticmethod
    def query_device_participant_for_participant_device(device_id: int, participant_id: int,
                                                        with_deleted: bool = False):
        return TeraDeviceParticipant.query.filter_by(id_device=device_id, id_participant=participant_id)\
            .execution_options(include_deleted=with_deleted).first()

    @classmethod
    def update(cls, update_id: int, values: dict):
        return

    @classmethod
    def insert(cls, dp):
        # Check if that the participant is associated with the same project as the device
        from opentera.db.models.TeraDeviceProject import TeraDeviceProject
        from opentera.db.models.TeraParticipant import TeraParticipant

        participant = TeraParticipant.get_participant_by_id(dp.id_participant)
        if not participant:
            raise IntegrityError(params='Participant not found',
                                 orig='TeraDeviceParticipant.insert', statement='insert')

        device_project = TeraDeviceProject.get_device_project_id_for_device_and_project(dp.id_device,
                                                                                        participant.id_project)

        if not device_project:
            raise IntegrityError(params='Device not associated to project',
                                 orig='TeraDeviceParticipant.insert', statement='insert')
        inserted_obj = super().insert(dp)
        return inserted_obj



from opentera.db.Base import BaseModel
from opentera.db.SoftDeleteMixin import SoftDeleteMixin
from opentera.db.SoftInsertMixin import SoftInsertMixin
from opentera.db.models.TeraTest import TeraTest
from opentera.db.models.TeraSession import TeraSession
from sqlalchemy import Column, ForeignKey, Integer, Sequence
from sqlalchemy.orm import relationship
from sqlalchemy.exc import IntegrityError


class TeraTestTypeProject(BaseModel, SoftDeleteMixin, SoftInsertMixin):
    __tablename__ = 't_tests_types_projects'
    id_test_type_project = Column(Integer, Sequence('id_test_type_project_sequence'), primary_key=True,
                                  autoincrement=True)
    id_test_type = Column('id_test_type', Integer, ForeignKey('t_tests_types.id_test_type', ondelete='cascade'),
                          nullable=False)
    id_project = Column('id_project', Integer, ForeignKey('t_projects.id_project', ondelete='cascade'), nullable=False)

    test_type_project_test_type = relationship("TeraTestType", viewonly=True)
    test_type_project_project = relationship("TeraProject", viewonly=True)

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []

        ignore_fields.extend(['test_type_project_test_type', 'test_type_project_project'])

        if minimal:
            ignore_fields.extend([])

        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def create_defaults(test=False):
        if test:
            from opentera.db.models.TeraTestType import TeraTestType
            from opentera.db.models.TeraProject import TeraProject

            pre_test = TeraTestType.get_test_type_by_id(1)
            post_test = TeraTestType.get_test_type_by_id(2)
            general_test = TeraTestType.get_test_type_by_id(3)

            project = TeraProject.get_project_by_projectname('Default Project #1')
            secret_project = TeraProject.get_project_by_projectname('Secret Project #1')

            ttp = TeraTestTypeProject()
            ttp.id_test_type = pre_test.id_test_type
            ttp.id_project = project.id_project
            TeraTestTypeProject.db().session.add(ttp)

            ttp = TeraTestTypeProject()
            ttp.id_test_type = post_test.id_test_type
            ttp.id_project = project.id_project
            TeraTestTypeProject.db().session.add(ttp)

            ttp = TeraTestTypeProject()
            ttp.id_test_type = general_test.id_test_type
            ttp.id_project = secret_project.id_project
            TeraTestTypeProject.db().session.add(ttp)

            TeraTestTypeProject.db().session.commit()

    @staticmethod
    def get_test_type_project_by_id(stp_id: int, with_deleted: bool = False):
        return TeraTestTypeProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_test_type_project=stp_id).first()

    @staticmethod
    def get_projects_for_test_type(test_type_id: int, with_deleted: bool = False):
        return TeraTestTypeProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_test_type=test_type_id).all()

    @staticmethod
    def get_tests_types_for_project(project_id: int, with_deleted: bool = False):
        return TeraTestTypeProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_project=project_id).all()

    @staticmethod
    def get_test_type_project_for_test_type_project(project_id: int, test_type_id: int, with_deleted: bool = False):
        return TeraTestTypeProject.query.execution_options(include_deleted=with_deleted)\
            .filter_by(id_project=project_id, id_test_type=test_type_id).first()

    @staticmethod
    def get_test_type_project_for_project_and_service(project_id: int, service_id: int, with_deleted: bool = False):
        from opentera.db.models.TeraTestType import TeraTestType
        return TeraTestTypeProject.query.execution_options(include_deleted=with_deleted).join(TeraTestType).\
            filter(TeraTestType.id_service == service_id).\
            filter(TeraTestTypeProject.id_project == project_id).all()

    @staticmethod
    def delete_with_ids(test_type_id: int, project_id: int, autocommit: bool = True):
        delete_obj: TeraTestTypeProject = TeraTestTypeProject.query.filter_by(id_test_type=test_type_id,
                                                                              id_project=project_id).first()
        if delete_obj:
            TeraTestTypeProject.delete(delete_obj.id_test_type_project, autocommit=autocommit)

    @staticmethod
    def check_integrity(obj_to_check):
        # Make sure that the service is associated to that project
        service_projects = [proj.id_project for proj in
                            obj_to_check.test_type_project_test_type.test_type_service.service_projects]
        if obj_to_check.id_project not in service_projects:
            # We must also associate that service to that project!
            from opentera.db.models.TeraServiceProject import TeraServiceProject
            new_service_project = TeraServiceProject()
            new_service_project.id_service = obj_to_check.test_type_project_test_type \
                .test_type_service.id_service
            new_service_project.id_project = obj_to_check.test_type_project_project.id_project
            TeraServiceProject.insert(new_service_project)

    @classmethod
    def insert(cls, ttp):
        # Check if that site of that project has the site associated to it
        from opentera.db.models.TeraTestTypeSite import TeraTestTypeSite
        from opentera.db.models.TeraProject import TeraProject
        project = TeraProject.get_project_by_id(project_id=ttp.id_project)
        tt_site = TeraTestTypeSite.get_test_type_site_for_test_type_and_site(site_id=project.id_site,
                                                                             test_type_id=ttp.id_test_type)
        if not tt_site:
            raise IntegrityError(params='Test type not associated to project site',
                                 orig='TeraTestTypeProject.insert', statement='insert')
        inserted_obj = super().insert(ttp)
        TeraTestTypeProject.check_integrity(inserted_obj)
        return inserted_obj

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        session_ids = [session.id_session for session in TeraSession.get_sessions_for_project(self.id_project,
                                                                                              with_deleted=with_deleted)]
        test_count = TeraTest.query.filter(TeraTest.id_session.in_(session_ids))\
            .filter(TeraTest.id_test_type == self.id_test_type).execution_options(include_deleted=with_deleted).count()

        if test_count > 0:
            return IntegrityError('Test type has tests in this project', self.id_test_type, 't_tests')
        return None

    @classmethod
    def update(cls, update_id: int, values: dict):
        return



from opentera.forms.TeraForm import *
from flask_babel import gettext
from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess
from opentera.db.models.TeraSession import TeraSessionStatus, TeraSession
from opentera.db.models.TeraProject import TeraProject
from opentera.db.models.TeraDeviceProject import TeraDeviceProject
from opentera.db.models.TeraUser import TeraUser


class TeraSessionForm:

    @staticmethod
    def get_session_form(user_access: DBManagerTeraUserAccess, specific_session_id: int = None, project_id: int = None):
        form = TeraForm("session")

        # If not allowed to access that session or new session, will return all accessibles lists
        if specific_session_id and not user_access.query_session(specific_session_id):
            specific_session_id = None

        session_info = None
        project_info = None
        if specific_session_id:
            session_info = TeraSession.get_session_by_id(specific_session_id)
            if session_info:
                project_info = TeraProject.get_project_by_id(session_info.get_associated_project_id())
        elif project_id:
            if project_id in user_access.get_accessible_projects_ids():
                project_info = TeraProject.get_project_by_id(project_id=project_id)

        # Building lists
        # Session types
        if not project_info:
            ses_types = user_access.get_accessible_session_types()
        else:
            ses_types_projects = user_access.query_session_types_for_project(project_info.id_project)
            ses_types = [ses_type.session_type_project_session_type for ses_type in ses_types_projects]

        if specific_session_id:
            # Make sure that the session type is there even if it's not related to that project anymore. This could
            # happen if a session type was removed from that project after the session was created
            if session_info.id_session_type not in [ses_type.id_session_type for ses_type in ses_types]:
                ses_types.append(session_info.session_session_type)

        st_list = []
        for st in ses_types:
            st_list.append(TeraFormValue(value_id=st.id_session_type, value=st.session_type_name))

        # Users
        if not project_info:
            users = user_access.get_accessible_users()
        else:
            users = project_info.get_users_in_project(include_superadmins=True, include_site_access=True)

        users_list = []
        users_ids = []
        for user in users:
            users_list.append(TeraFormValue(value_id=user.id_user, value=user.get_fullname()))
            users_ids.append(user.id_user)

        # Also appends super-admins if user is super-admin
        if user_access.user.user_superadmin:
            for user in TeraUser.get_superadmins():
                if user.id_user not in users_ids:
                    users_list.append(TeraFormValue(value_id=user.id_user, value=user.get_fullname()))
                    users_ids.append(user.id_user)

        # Always append querying user if not there
        if user_access.user.id_user not in users_ids:
            users_list.append(TeraFormValue(value_id=user_access.user.id_user, value=user_access.user.get_fullname()))
            users_ids.append(user_access.user.id_user)

        # Devices
        if not project_info:
            devices = user_access.get_accessible_devices()
        else:
            devices = [dev.device_project_device for dev in
                       TeraDeviceProject.get_devices_for_project(project_info.id_project)]

        devices_list = list()
        for device in devices:
            devices_list.append(TeraFormValue(value_id=device.id_device, value=device.device_name))

        # Participants
        if not project_info:
            participants = user_access.get_accessible_participants()
        else:
            participants = user_access.query_all_participants_for_project(project_info.id_project)
        parts_list = list()
        for part in participants:
            parts_list.append(TeraFormValue(value_id=part.id_participant, value=part.participant_name))

        # Services
        if not project_info:
            services = user_access.get_accessible_services()
        else:
            services = user_access.query_services_for_project(project_info.id_project)
        services_list = list()
        for service in services:
            services_list.append(TeraFormValue(value_id=service.id_service, value=service.service_name))

        # Session status
        status_list = []
        for status in TeraSessionStatus:
            # Translation is done here, since we have to proper language context
            status_name = gettext('Unknown')
            if status.value == TeraSessionStatus.STATUS_CANCELLED.value:
                status_name = gettext('Cancelled')
            if status.value == TeraSessionStatus.STATUS_COMPLETED.value:
                status_name = gettext('Completed')
            if status.value == TeraSessionStatus.STATUS_INPROGRESS.value:
                status_name = gettext('In progess')
            if status.value == TeraSessionStatus.STATUS_NOTSTARTED.value:
                status_name = gettext('Planned')
            if status.value == TeraSessionStatus.STATUS_TERMINATED.value:
                status_name = gettext('Interrupted')

            status_list.append(TeraFormValue(value_id=status.value, value=status_name))

        # Sections
        section = TeraFormSection("informations", gettext("Information"))
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_session", gettext("Session ID"), "hidden", True))
        section.add_item(TeraFormItem("session_uuid", gettext("Session UUID"), "hidden", True))
        section.add_item(TeraFormItem("session_name", gettext("Session Name"), "text", True))
        section.add_item(TeraFormItem("id_session_type", gettext("Session Type"), "array", True, item_values=st_list))
        section.add_item(TeraFormItem('session_creator_user', gettext('Session User Creator'), 'hidden', False))
        section.add_item(TeraFormItem("id_creator_user", gettext("Session User Creator"), "array", True,
                                      item_values=users_list, item_options={"readonly": True},
                                      item_condition=TeraFormItemCondition(condition_item='session_creator_user',
                                                                           condition_operator='NOT NULL',
                                                                           condition_condition=None)))
        section.add_item(TeraFormItem('session_creator_device', gettext('Session Creator Device'), 'hidden', False))
        section.add_item(TeraFormItem("id_creator_device", gettext("Session Creator Device ID"), "array", True,
                                      item_values=devices_list, item_options={"readonly": True},
                                      item_condition=TeraFormItemCondition(condition_item='session_creator_device',
                                                                           condition_operator='NOT NULL',
                                                                           condition_condition=None)
                                      ))
        section.add_item(TeraFormItem('session_creator_participant', gettext('Session Creator Participant'), 'hidden',
                                      False))
        section.add_item(TeraFormItem("id_creator_participant", gettext("Session Creator Participant"), "array", True,
                                      item_values=parts_list, item_options={"readonly": True},
                                      item_condition=TeraFormItemCondition(condition_item='session_creator_participant',
                                                                           condition_operator='NOT NULL',
                                                                           condition_condition=None)))
        section.add_item(TeraFormItem('session_creator_service', gettext('Session Creator Service'), 'hidden',
                                      False))
        section.add_item(TeraFormItem("id_creator_service", gettext("Session Creator Service"), "array", True,
                                      item_values=services_list, item_options={"readonly": True},
                                      item_condition=TeraFormItemCondition(condition_item='session_creator_service',
                                                                           condition_operator='NOT NULL',
                                                                           condition_condition=None)))
        section.add_item(TeraFormItem("session_start_datetime", gettext("Start Date"), "datetime", True))
        section.add_item(TeraFormItem("session_duration", gettext("Duration"), "duration", True,
                                      item_options={"default": 0, "readonly": True}))
        # Session status is hidden as it needs to be handled elsewhere for now
        section.add_item(TeraFormItem("session_status", gettext("State"), "array", True, item_values=status_list))
        section.add_item(TeraFormItem("session_parameters", gettext("Parameters"), "json", False))
        section.add_item(TeraFormItem("session_comments", gettext("Comments"), "longtext", False))

        # Hidden as handled elsewhere
        # section.add_item(TeraFormItem("session_participants", gettext("Participants"), "hidden", False))
        # section.add_item(TeraFormItem("session_users", gettext("Users"), "hidden", False))
        # section.add_item(TeraFormItem("session_devices", gettext("Devices"), "hidden", False))
        section.add_item(TeraFormItem("session_has_device_data", gettext("Session Has Device Data"), "hidden", False))

        return form.to_dict()



from opentera.forms.TeraForm import *
from flask_babel import gettext
from modules.DatabaseModule.DBManagerTeraUserAccess import DBManagerTeraUserAccess


class TeraTestTypeForm:

    @staticmethod
    def get_test_type_form(services):
        form = TeraForm("test_type")

        # Building lists
        services_list = []
        for service in services:
            services_list.append(TeraFormValue(value_id=service.id_service, value=service.service_name))

        # Sections
        section = TeraFormSection("informations", "")
        form.add_section(section)

        # Items
        section.add_item(TeraFormItem("id_test_type", gettext("Test Type ID"), "hidden", True))
        section.add_item(TeraFormItem("test_type_name", gettext("Name"), "text", True))
        section.add_item(TeraFormItem("id_service", gettext("Service"), "array", item_required=True,
                                      item_values=services_list)
                         )
        section.add_item(TeraFormItem("test_type_uuid", gettext("UUID"), "hidden", True))
        section.add_item(TeraFormItem("test_type_key", gettext("Unique Key"), "text", False))
        section.add_item(TeraFormItem("test_type_has_json_format", gettext("Expose JSON structure"), "boolean", False))
        section.add_item(TeraFormItem("test_type_has_web_format", gettext("Expose Web interface"), "boolean", False))
        section.add_item(TeraFormItem("test_type_has_web_editor", gettext("Expose Web editor"), "boolean", False))
        section.add_item(TeraFormItem("test_type_description", gettext("Description"), "longtext", False))

        return form.to_dict()



import time
import inspect


def timing(f):
    def wrap(*args, **kwargs):
        time1 = time.time()
        ret = f(*args, **kwargs)
        time2 = time.time()
        exec_time = (time2-time1)*1000.0
        if exec_time > 200:
            print("WARNING: %s: %s took %0.3f ms" % (f.__module__, f.__name__, exec_time))
        return ret
    return wrap


# Will not play well with @staticmethod and @classmethod for now...
def timing_for_all_methods(cls):
    for name, method in inspect.getmembers(cls):
        if (not inspect.ismethod(method) and not inspect.isfunction(method)) or inspect.isbuiltin(method) \
                or method.__name__ == '__init__':
            continue
        # print("Decorating function %s" % name)
        setattr(cls, name, timing(method))
    return cls



from opentera.utils.assets.AssetFile import AssetFile
from dataclasses import dataclass
from datetime import timedelta


@dataclass
class AssetVideoFile(AssetFile):
    video_duration: timedelta = timedelta(0)



from opentera.db.Base import BaseModel
from sqlalchemy import Column, ForeignKey, Integer, String, Sequence, Boolean, TIMESTAMP
from sqlalchemy.orm import relationship


class TeraUserPreference(BaseModel):
    __tablename__ = 't_users_preferences'
    id_user_preference = Column(Integer, Sequence('id_userpreference_sequence'), primary_key=True, autoincrement=True)
    id_user = Column(Integer, ForeignKey('t_users.id_user', ondelete='cascade'), nullable=False)
    user_preference_app_tag = Column(String, nullable=False)
    user_preference_preference = Column(String, nullable=False)

    user_preference_user = relationship("TeraUser")

    def to_json(self, ignore_fields=None, minimal=False):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.extend(['user_preference_user'])
        if minimal:
            ignore_fields.extend([])
        rval = super().to_json(ignore_fields=ignore_fields)

        return rval

    @staticmethod
    def get_user_preferences_for_app(app_tag: str):
        return TeraUserPreference.query.filter_by(user_preference_app_tag=app_tag).first()

    @staticmethod
    def get_user_preference_by_id(user_pref_id: int):
        return TeraUserPreference.query.filter_by(id_user_prefrence=user_pref_id).first()

    @staticmethod
    def get_user_preferences_for_user(user_id: int):
        return TeraUserPreference.query.filter_by(id_user=user_id).all()

    @staticmethod
    def get_user_preferences_for_user_and_app(user_id: int, app_tag: str):
        return TeraUserPreference.query.filter_by(id_user=user_id, user_preference_app_tag=app_tag).first()

    @staticmethod
    def create_defaults(test=False):

        if test:
            from opentera.db.models.TeraUser import TeraUser
            super_admin = TeraUser.get_user_by_username('admin')
            site_admin = TeraUser.get_user_by_username('siteadmin')

            new_pref = TeraUserPreference()
            new_pref.id_user = super_admin.id_user
            new_pref.user_preference_app_tag = 'openteraplus'
            new_pref.user_preference_preference = '{"language": "fr", "notification_sounds": true}'
            TeraUserPreference.db().session.add(new_pref)

            new_pref = TeraUserPreference()
            new_pref.id_user = site_admin.id_user
            new_pref.user_preference_app_tag = 'openteraplus'
            new_pref.user_preference_preference = '{"language": "en", "notification_sounds": false}'
            TeraUserPreference.db().session.add(new_pref)

            new_pref = TeraUserPreference()
            new_pref.id_user = super_admin.id_user
            new_pref.user_preference_app_tag = 'anotherapp'
            new_pref.user_preference_preference = '{"gui_style": 1, "auto_save": false}'
            TeraUserPreference.db().session.add(new_pref)

    @staticmethod
    def insert_or_update_or_delete_user_preference(user_id: int, app_tag: str, prefs: str):
        if prefs:
            # Check if prefs is a valid json structure
            import json
            if not isinstance(prefs, dict):
                try:
                    json.loads(prefs)
                except ValueError as err:
                    raise err
            else:
                try:
                    prefs = json.dumps(prefs)
                except ValueError as err:
                    raise err

        # Check if we have an existing preference for that user
        existing_pref = TeraUserPreference.get_user_preferences_for_user_and_app(user_id=user_id, app_tag=app_tag)

        if existing_pref:
            # Update or delete
            if prefs is None or prefs == '':
                TeraUserPreference.db().session.delete(existing_pref)
            else:
                # Update pref
                existing_pref.user_preference_preference = prefs
        else:
            # Insert
            new_user_pref = TeraUserPreference()
            new_user_pref.id_user = user_id
            new_user_pref.user_preference_app_tag = app_tag
            new_user_pref.user_preference_preference = prefs
            TeraUserPreference.db().session.add(new_user_pref)
        TeraUserPreference.db().session.commit()



from opentera.redis.RedisClient import RedisClient
from enum import Enum, unique
import opentera.messages.python as messages
from opentera.logging.LoggingClient import LoggingClient

import json
import datetime


@unique
class ModuleNames(Enum):
    FLASK_MODULE_NAME = str("TeraServer.FlaskModule")
    TWISTED_MODULE_NAME = str("TeraServer.TwistedModule")
    LOGIN_MODULE_NAME = str("TeraServer.LoginModule")
    USER_MANAGER_MODULE_NAME = str("TeraServer.UserManagerModule")
    DATABASE_MODULE_NAME = str("TeraServer.DatabaseModule")
    SERVICE_LAUNCHER_NAME = str("TeraServer.ServiceLauncherModule")


def create_module_message_topic_from_name(name: ModuleNames):
    return 'module.' + name.value + '.messages'


def create_module_event_topic_from_name(name: ModuleNames, subtopic=None):
    if subtopic:
        return 'module.' + name.value + '.events.' + str(subtopic)
    else:
        return 'module.' + name.value + '.events'


class BaseModule(RedisClient):
    """
        BaseModule will handle basic registration of topics and events.

    """
    def __init__(self, module_name, config):

        # Set module name
        # TODO verify module name to be unique
        self.module_name = module_name

        # Store config
        self.config = config

        # Store RPC API
        self.rpc_api = dict()

        # Logger
        self.logger = LoggingClient(config.redis_config, 'LoggingClient_' + self.__class__.__name__)

        # Init redis with configuration
        RedisClient.__init__(self, config=config.redis_config)

    def get_name(self):
        return self.module_name

    def event_topic_name(self):
        return 'module.' + self.module_name + '.events'

    def send_event_message(self, event, topic: str):
        message = self.create_event_message(topic)
        any_message = messages.Any()
        any_message.Pack(event)
        message.events.extend([any_message])
        self.publish(message.header.topic, message.SerializeToString())

    def send_module_message(self, message: messages.TeraModuleMessage):
        self.publish(message.head.dest, message.SerializeToString())

    def redisConnectionMade(self):
        print('*************************** BaseModule.connectionMade', self.module_name)

        # Build RPC interface
        self.setup_rpc_interface()

        # Build standard interface
        self.build_interface()

        # Setup pubsub for module, needs to be overridden
        self.setup_module_pubsub()

    def setup_module_pubsub(self):
        pass

    def build_interface(self):
        # TeraModuleMessage Interface
        def messages_subscribed_callback(*args):
            print('messages_subscribed_callback for ', self.module_name)

        ret1 = self.subscribe_pattern_with_callback('module.' + self.module_name + '.messages', self.notify_module_messages)
        ret1.addCallback(messages_subscribed_callback)

        # RPC messages
        def rpc_subscribed_callback(*args):
            print('rpc_subscribed_callback for ', self.module_name)
        ret2 = self.subscribe_pattern_with_callback('module.' + self.module_name + '.rpc', self.notify_module_rpc)
        ret2.addCallback(rpc_subscribed_callback)

    def setup_rpc_interface(self):
        pass

    def notify_module_messages(self, pattern, channel, message):
        """
        We have received a published message from redis
        """
        print('BaseModule - Received message', self, pattern, channel, message)
        pass

    def notify_module_rpc(self, pattern, channel, message):
        # print('BaseModule - Received rpc', self, pattern, channel, message, ' thread:', threading.current_thread())

        try:
            # Look for a RPCMessage
            rpc_message = messages.RPCMessage()
            rpc_message.ParseFromString(message)

            if self.rpc_api.__contains__(rpc_message.method):

                # RPC method found, call it with the args
                args = list()
                kwargs = dict()

                # TODO type checking with declared rpc interface ?
                for value in rpc_message.args:
                    # Append the oneof value to args
                    args.append(getattr(value, value.WhichOneof('arg_value')))

                # Call callback function
                ret_value = self.rpc_api[rpc_message.method]['callback'](*args, **kwargs)

                # More than we need?
                my_dict = {'method': rpc_message.method,
                           'id': rpc_message.id,
                           'pattern': pattern,
                           'status': 'OK',
                           'return_value': ret_value}

                json_data = json.dumps(my_dict)

                # Return result (a json string)
                self.publish(rpc_message.reply_to, json_data)

        except:
            import sys
            print('Error calling rpc method', message, sys.exc_info())
            my_dict = {'method': rpc_message.method,
                       'id': rpc_message.id,
                       'pattern': pattern,
                       'status': 'Error',
                       'return_value': None}

            json_data = json.dumps(my_dict)

            # Return result (a json string)
            self.publish(rpc_message.reply_to, json_data)

    def create_tera_message(self, dest='', seq=0):
        tera_message = messages.TeraModuleMessage()
        tera_message.head.version = 1
        tera_message.head.time = datetime.datetime.now().timestamp()
        tera_message.head.seq = seq
        tera_message.head.source = 'module.' + self.module_name
        tera_message.head.dest = dest
        return tera_message

    def create_event_message(self, topic):
        event_message = messages.TeraEvent()
        event_message.header.version = 1
        event_message.header.time = datetime.datetime.now().timestamp()
        event_message.header.topic = topic
        return event_message

    def create_module_message(self, src, dest):
        """
            message Header {
            uint32 version = 1;
            double time = 2;
            uint32 seq = 3;
            string source = 4;
            string dest = 5;
        }
        Header head = 1;
        repeated google.protobuf.Any data = 2;
        """

        module_message = messages.TeraModuleMessage()
        module_message.head.version = 1
        module_message.head.time = datetime.datetime.now().timestamp()
        module_message.head.seq = 0
        module_message.head.source = src
        module_message.head.dest = dest
        return module_message

    def source_name(self):
        return "module." + self.module_name + ".messages"

    def _send_disconnect_module_message(self, uuid: str):
        module_message = self.create_module_message(
            src='module.' + self.module_name + '.messages',
            dest=create_module_event_topic_from_name(ModuleNames.TWISTED_MODULE_NAME, uuid)
        )
        server_command = messages.ServerCommand()
        server_command.type = messages.ServerCommand.CMD_CLIENT_DISCONNECT
        command_message = messages.Any()
        command_message.Pack(server_command)
        module_message.data.extend([command_message])
        self.send_module_message(module_message)

    def send_user_disconnect_module_message(self, user_uuid: str):
        self._send_disconnect_module_message(user_uuid)

    def send_participant_disconnect_module_message(self, participant_uuid: str):
        self._send_disconnect_module_message(participant_uuid)

    def send_device_disconnect_module_message(self, device_uuid: str):
        self._send_disconnect_module_message(device_uuid)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ServerCommand.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13ServerCommand.proto\x12\x11opentera.protobuf\x1a\x19google/protobuf/any.proto\"\xc3\x01\n\rServerCommand\x12:\n\x04type\x18\x01 \x01(\x0e\x32,.opentera.protobuf.ServerCommand.CommandType\x12\"\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\"R\n\x0b\x43ommandType\x12\x0f\n\x0b\x43MD_UNKNOWN\x10\x00\x12\x17\n\x13\x43MD_REGISTER_DEVICE\x10\x01\x12\x19\n\x15\x43MD_CLIENT_DISCONNECT\x10\x02\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ServerCommand_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SERVERCOMMAND._serialized_start=70
  _SERVERCOMMAND._serialized_end=265
  _SERVERCOMMAND_COMMANDTYPE._serialized_start=183
  _SERVERCOMMAND_COMMANDTYPE._serialized_end=265
# @@protoc_insertion_point(module_scope)



import inspect
import datetime
import time
import typing as t
import sqlalchemy.sql.sqltypes
from flask_sqlalchemy import SQLAlchemy, query, model
from sqlalchemy import Column, BigInteger
from sqlalchemy.inspection import inspect as sqlinspector
from sqlalchemy.exc import IntegrityError, SQLAlchemyError
from sqlalchemy.orm import Session


class _QueryProperty:
    def __get__(self, obj: model.Model | None, cls: t.Type[model.Model]) -> query.Query:
        return cls.db().session.query(cls)


class BaseMixin(object):
    version_id = Column(BigInteger, nullable=False, default=int(time.time()*1000))

    # Using timestamp as version tracker - multiplying by 1000 to keep ms part without using floats (which seems to
    # cause problems with the mapper)
    __mapper_args__ = {
        'version_id_col': version_id,
        'version_id_generator': lambda version: int(time.time()*1000)
    }

    # This needs to be initialized by app
    __db__: SQLAlchemy = None

    @classmethod
    def set_db(cls, db: SQLAlchemy):
        cls.__db__ = db

    @classmethod
    def create_all(cls):
        if cls.__db__ and cls.__db__.engine:
            cls.__db__.create_all()
            cls.metadata.create_all(cls.__db__.engine)

    query = _QueryProperty()

    @classmethod
    def db(cls) -> SQLAlchemy:
        return cls.__db__

    def to_json(self, ignore_fields=None):
        if ignore_fields is None:
            ignore_fields = []
        pr = {}
        # Add relationships in ignore_fields by default
        for relation in sqlinspector(self.__class__).relationships.items():
            if relation[0] not in ignore_fields:
                # print('-- Ignoring ' + relation[0] + ' in to_json()')
                ignore_fields.append(relation[0])

        for name in dir(self):
            if (self.is_valid_property_name(name) and name not in ignore_fields) or name == 'deleted_at':
                value = getattr(self, name)
                if name == 'deleted_at' and value is None:
                    continue  # If deleted field, but not deleted, don't add to the json
                if self.is_valid_property_value(value):
                    if isinstance(value, datetime.datetime) or isinstance(value, datetime.date):
                        value = value.isoformat()
                    if isinstance(value, datetime.timedelta):
                        # Strip too many zeros at the end
                        value_times = str(value).split(".")
                        if len(value_times) > 1:
                            value_times[1] = value_times[1][0:3]
                            value = value_times[0] + '.' + value_times[1]
                        else:
                            value = value_times[0]
                    pr[name] = value
        return pr

    def from_json(self, json, ignore_fields=None):
        if ignore_fields is None:
            ignore_fields = []
        ignore_fields.append('deleted_at')
        for name in json:
            if name not in ignore_fields:
                if hasattr(self, name):
                    # Test for datetime as string
                    # This is a fix for SQLITE that does not convert str to datetime automatically
                    if isinstance(self.__table__.columns[name].type, sqlalchemy.sql.sqltypes.TIMESTAMP) \
                            and isinstance(json[name], str) and len(json[name]) > 0:
                        setattr(self, name, datetime.datetime.fromisoformat(json[name]))
                    else:
                        setattr(self, name, json[name])
                else:
                    print('Attribute ' + name + ' not found.')

    def to_json_create_event(self):
        # Default is None, will not be sent
        return None

    def to_json_update_event(self):
        # Default is None, will not be sent
        return None

    def to_json_delete_event(self):
        # Default is None, will not be sent
        return None

    @staticmethod
    def is_valid_property_name(name: str) -> bool:
        return not name.startswith('__') and not name.startswith('_') and not name.startswith('query') and \
               not name.startswith('metadata') and name != 'version_id' and name != 'registry' and name != 'deleted_at'

    @staticmethod
    def is_valid_property_value(value: str) -> bool:
        return not inspect.ismethod(value) and not inspect.isfunction(value)

    @classmethod
    def clean_values(cls, values: dict):
        # This method is used to remove item from the values dict that are not properties of the object
        obj_properties = list()

        # Build available properties
        for name in dir(cls):
            value = getattr(cls, name)
            if cls.is_valid_property_name(name) and cls.is_valid_property_value(value):
                obj_properties.append(name)

        # Remove any property not in the available list
        clean_values = values.copy()
        for value in values:
            if value not in obj_properties:
                del clean_values[value]

        return clean_values

    @classmethod
    def get_count(cls, filters: dict = None, with_deleted: bool = False) -> int:
        count_query = cls.db().session.query(cls).execution_options(include_deleted=with_deleted)
        if filters:
            count_query = count_query.filter_by(**filters)
        return count_query.count()

    @classmethod
    def get_primary_key_name(cls) -> str:
        from sqlalchemy import inspect
        return inspect(cls).primary_key[0].name

    @classmethod
    def get_model_name(cls) -> str:
        key = cls.get_primary_key_name()
        key_split = key.split('_', 1)
        return key_split[-1]

    @classmethod
    def update(cls, update_id: int, values: dict):
        update_values = cls.clean_values(values)
        if len(update_values) != len(values):
            raise SQLAlchemyError('Invalid values passed to update')

        if cls.get_primary_key_name() in update_values and update_id != update_values[cls.get_primary_key_name()]:
            raise SQLAlchemyError(f'Primary key cannot be updated ({cls.get_primary_key_name()})')

        # with Session(cls.db().engine) as session:
        update_obj = cls.db().session.query(cls).filter(getattr(cls, cls.get_primary_key_name()) == update_id).first()

        if update_obj:
            update_obj.from_json(update_values)
            cls.db().session.commit()
        else:
            raise SQLAlchemyError(cls.__name__ + ' with id ' + str(update_id) + ' cannot update.')

    @classmethod
    def commit(cls):
        cls.db().session.commit()

    @classmethod
    def insert(cls, db_object):
        if getattr(db_object, 'soft_insert', None):
            return db_object.soft_insert(db_object)
        else:
            # Clear primary key value
            setattr(db_object, cls.get_primary_key_name(), None)

            # Add to database session and commit
            cls.db().session.add(db_object)
            cls.commit()
            return db_object

    def delete_check_integrity(self, with_deleted: bool = False) -> IntegrityError | None:
        return None  # Can delete by default

    @classmethod
    def delete(cls, id_todel, autocommit: bool = True, hard_delete: bool = False):
        if hard_delete:
            cls.handle_include_deleted_flag(True)
            autocommit = False
        delete_obj = cls.db().session.query(cls).filter(getattr(cls, cls.get_primary_key_name()) == id_todel).first()

        if delete_obj:
            has_soft_delete = getattr(delete_obj, 'soft_delete', None) is not None
            has_hard_delete = getattr(delete_obj, 'hard_delete', None) is not None

            if (has_soft_delete and not delete_obj.deleted_at) or not has_soft_delete:
                # Don't check integrity for already soft-deleted objects
                cannot_be_deleted_exception = delete_obj.delete_check_integrity()
                if cannot_be_deleted_exception:
                    raise cannot_be_deleted_exception

            if has_soft_delete and not hard_delete:
                delete_obj.soft_delete()
            else:
                # if has_soft_delete:
                #     # Check that object was soft deleted before doing a hard delete
                #     if not delete_obj.deleted_at:
                #         # Object must be soft deleted first before being hard deleted!
                #         raise SQLAlchemyError(cls.__name__ + ' with id ' + str(id_todel) +
                #                               ' cannot be hard deleted: not soft deleted beforehand!')
                if has_hard_delete and hard_delete:
                    delete_obj.hard_delete()
                else:
                    cls.db().session.delete(delete_obj)
            if autocommit:
                cls.commit()
        else:
            raise SQLAlchemyError(cls.__name__ + ' with id ' + str(id_todel) + ' cannot delete.')
        if hard_delete:
            cls.handle_include_deleted_flag(False)

    @classmethod
    def undelete(cls, id_to_undelete):
        undelete_obj = cls.db().session.query(cls).execution_options(include_deleted=True).filter(
            getattr(cls, cls.get_primary_key_name()) == id_to_undelete).first()
        if undelete_obj:
            if getattr(undelete_obj, 'soft_undelete', None):
                undelete_obj.soft_undelete()
            cls.commit()
        else:
            print(cls.__name__ + ' with id ' + str(id_to_undelete) + ' cannot undelete.')
            raise SQLAlchemyError(cls.__name__ + ' with id ' + str(id_to_undelete) + ' cannot undelete.')

    @classmethod
    def handle_include_deleted_flag(cls, include_deleted=False):
        if 'include_deleted' not in cls.db().session.info:
            cls.db().session.info['include_deleted'] = list()

        if include_deleted:
            cls.db().session.info['include_deleted'].append(cls.get_model_name())
        else:
            cls.db().session.info['include_deleted'].pop(-1)

    @classmethod
    def query_with_filters(cls, filters=None, with_deleted: bool = False):
        if filters is None:
            filters = dict()

        return cls.db().session.query(cls).execution_options(include_deleted=with_deleted).filter_by(**filters).all()

    @classmethod
    def get_json_schema(cls) -> dict:
        # Get model prefix (name)
        model_name = cls.get_model_name()

        # Browse each
        pr_dict = dict()
        for name in dir(cls):
            value = getattr(cls, name)
            if cls.is_valid_property_name(name) and cls.is_valid_property_value(value) and \
                    (name.startswith(model_name) or name.startswith('id')):
                # Ok so far, do we have a column and not a relationship or something else?
                if 'ColumnProperty' in str(type(value.prop)):
                    # Get correct data type
                    data_type = 'object'
                    data_format = None
                    column_type = str(value.prop.columns[0].type).lower()
                    default_value = value.prop.columns[0].default
                    if 'string' in column_type or 'timestamp' in column_type or 'varchar' in column_type:
                        data_type = 'string'
                        if 'uuid' in name:
                            data_format = 'uuid'
                        if 'timestamp' in column_type:
                            data_format = 'date-time'
                    if 'integer' in column_type:
                        data_type = 'integer'
                    if 'boolean' in column_type:
                        data_type = 'boolean'

                    pr_dict[name] = {'type': data_type, 'required': not value.prop.columns[0].nullable}
                    if data_format:
                        pr_dict[name]['format'] = data_format
                    if default_value:
                        if hasattr(default_value, 'arg'):
                            pr_dict[name]['default'] = default_value.arg

        schema = {model_name: {'properties': pr_dict, 'type': 'object'}}

        return schema

    @classmethod
    def validate_required_fields(cls, json_data: dict, ignore_fields: list = None):
        if not ignore_fields:
            ignore_fields = []

        # Get model prefix (name)
        model_name = cls.get_model_name()
        missing_fields = []

        # Browse each
        pr_dict = dict()
        for name in dir(cls):
            value = getattr(cls, name)
            if cls.is_valid_property_name(name) and cls.is_valid_property_value(value) and \
                    (name.startswith(model_name) or name.startswith('id')) and not name.endswith('uuid'):
                # Ok so far, check if column is required or not
                if 'ColumnProperty' in str(type(value.prop)):
                    if not value.prop.columns[0].nullable and name not in json_data and name not in ignore_fields and \
                            not value.prop.columns[0].default:
                        missing_fields.append(name)

        return missing_fields


# Declarative base, inherit from Base for all models
BaseModel = sqlalchemy.orm.declarative_base(cls=BaseMixin)



from abc import ABC, abstractmethod
import datetime
from dataclasses import dataclass


@dataclass
class BaseAsset(ABC):
    asset_name: str = ''
    asset_datetime: datetime = datetime.datetime.now()
    asset_uuid: str = ''

    @property
    @abstractmethod
    def asset_type(self) -> str:
        return 'unknown'

    def to_json(self) -> dict:
        attributes = [attr for attr in dir(self) if not callable(getattr(self, attr)) and not attr.startswith("_")]
        json = {k: v for k, v in zip([attr for attr in attributes], [getattr(self, attr) for attr in attributes])}

        return json

    def from_json(self, json: dict):
        for key in json.keys():
            if key in dir(self) and not callable(getattr(self, key)) and not key.startswith("_") \
                    and key != 'asset_type':
                setattr(self, key, json[key])




import uuid

from requests import Response
from flask import request


class TeraServiceClient:

    def __init__(self, token_dict: dict, token: str, config_man):
        self.__service_uuid = token_dict['service_uuid']
        self.__service_token = token

        backend_hostname = config_man.backend_config["hostname"]
        backend_port = str(config_man.backend_config["port"])

        self.__backend_url = 'https://' + backend_hostname + ':' + backend_port

    @property
    def service_uuid(self):
        return self.__service_uuid

    @service_uuid.setter
    def service_uuid(self, s_uuid: uuid):
        self.__service_uuid = s_uuid

    @property
    def service_token(self):
        return self.__service_token

    @service_token.setter
    def service_token(self, token: str):
        self.__service_token = token

    def do_get_request_to_backend(self, path: str) -> Response:
        from requests import get
        request_headers = {'Authorization': 'OpenTera ' + self.__service_token}
        # TODO: remove verify=False and check certificate
        backend_response = get(url=self.__backend_url + path, headers=request_headers, verify=False)
        return backend_response

    def get_service_infos(self) -> dict:
        response = self.do_get_request_to_backend('/api/service/services?uuid_service=' + self.__service_uuid)
        if response.status_code == 200:
            return response.json()[0]
        return {}

    def __repr__(self):
        return '<TeraServiceClient - UUID: ' + self.__service_uuid \
               + ', Token: ' + self.__service_token + '>'



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: UserEvent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fUserEvent.proto\x12\x11opentera.protobuf\"\x9a\x02\n\tUserEvent\x12\x11\n\tuser_uuid\x18\x01 \x01(\t\x12\x34\n\x04type\x18\x02 \x01(\x0e\x32&.opentera.protobuf.UserEvent.EventType\x12\x15\n\ruser_fullname\x18\x03 \x01(\t\"\xac\x01\n\tEventType\x12\x10\n\x0cUSER_UNKNOWN\x10\x00\x12\x12\n\x0eUSER_CONNECTED\x10\x01\x12\x15\n\x11USER_DISCONNECTED\x10\x02\x12\x10\n\x0cUSER_DELETED\x10\x03\x12\x0e\n\nUSER_ADDED\x10\x04\x12\x10\n\x0cUSER_UPDATED\x10\x05\x12\x17\n\x13USER_JOINED_SESSION\x10\x06\x12\x15\n\x11USER_LEFT_SESSION\x10\x07\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'UserEvent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _USEREVENT._serialized_start=39
  _USEREVENT._serialized_end=321
  _USEREVENT_EVENTTYPE._serialized_start=149
  _USEREVENT_EVENTTYPE._serialized_end=321
# @@protoc_insertion_point(module_scope)
